! ----------------------------------------------------------------------
! MESSAGE PASSING INTERFACE TEST CASE SUITE
! 
! Copyright - 1996 Intel Corporation
! 
! Intel Corporation hereby grants a non-exclusive license under Intel's
! copyright to copy, modify and distribute this software for any purpose 
! and without fee, provided that the above copyright notice and the
! following paragraphs appear on all copies.
! 
! Intel Corporation makes no representation that the test cases
! comprising this suite are correct or are an accurate representation
! of any standard.
! 
! IN NO EVENT SHALL INTEL HAVE ANY LIABILITY FOR ANY DIRECT, INDIRECT
! OR SPECULATIVE DAMAGES, (INCLUDING WITHOUT LIMITING THE FOREGOING,
! CONSEQUENTIAL, INCIDENTAL AND SPECIAL DAMAGES) INCLUDING, BUT NOT
! LIMITED TO INFRINGEMENT, LOSS OF USE, BUSINESS INTERRUPTIONS, AND
! LOSS OF PROFITS, IRRESPECTIVE OF WHETHER INTEL HAS ADVANCE NOTICE OF
! THE POSSIBILITY OF ANY SUCH DAMAGES.
! 
! INTEL CORPORATION SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING,
! BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
! FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  THE SOFTWARE PROVIDED
! HEREUNDER IS ON AN "AS IS" BASIS AND INTEL CORPORATION HAS NO
! OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS OR
! MODIFICATIONS.
! 
! ----------------------------------------------------------------------
! **********************************************************************
!                  Test for MPI_Errhandler_free
!
! This is the same as the test for MPI_Errhandler_set, except that
! MPI_Errhandler_free is called.  Note that the errorhandler should still
! be used while the communicator is in use.
! 
! MPI_COMM_WORLD is Dup'ed, and each are given a unique errorhandler.
! 
! An error is generated in both comms, plus one with MPI_COMM_NULL (which should
! correspond to MPI_COMM_WORLD's error handler).  This test verifies each
! handler is called correctly.
!
! Test history:
!    1  05/02/96     gt   Created
!
! **********************************************************************
#include "f90options.h"


      INTEGER FUNCTION ERR1(COMMIN, CODE)

      INCLUDE 'mpitestf90.h'
      INCLUDE 'externalf90.h'


      INTEGER COMMIN
      INTEGER CODE
      INTEGER IERR

      INTEGER PASS
      INTEGER FAIL
      INTEGER COUNT1
      INTEGER COUNT2
 
      INTEGER COMM

      COMMON /ERRCOM/PASS, FAIL, COUNT1, COUNT2, COMM

      INTEGER CLASS

      CHARACTER*(IOSIZE) INFOBUF
      CHARACTER*(MPI_MAX_ERROR_STRING) ERRORSTRING

!
!   Handles errors for MPI_COMM_WORLD
!
      IF (COUNT1 .EQ. 0) THEN
!         Bad length on MPI_SEND
        CALL MPI_ERROR_CLASS(CODE, CLASS, IERR)
        IF (CLASS .NE. MPI_ERR_COUNT) THEN
          INFOBUF='First error returned incorrect code'
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        ELSE
         PASS = PASS + 1
        END IF
      ELSE IF (COUNT1 .EQ. 1) THEN
!         Bad communicator in MPI_BARRIER
        CALL MPI_ERROR_CLASS(CODE, CLASS, IERR)
        IF (CLASS .NE. MPI_ERR_COMM) THEN
          INFOBUF='Second error returned incorrect code'
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        ELSE
         PASS = PASS + 1
        END IF
      ELSE
        INFOBUF='Errhandler 1 called too many times'
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        FAIL = FAIL + 1
      END IF

      IF (COMMIN .NE. MPI_COMM_WORLD) THEN
        INFOBUF='Errhandler 1 called w/ incorrect communicator'
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        FAIL = FAIL + 1
      END IF

      COUNT1 = COUNT1 + 1
      ERR1 = 0

      END



      INTEGER FUNCTION ERR2(COMMIN, CODE)

      INCLUDE 'mpitestf90.h'
      INCLUDE 'externalf90.h'


      INTEGER COMMIN
      INTEGER CODE

      INTEGER IERR
      INTEGER PASS
      INTEGER FAIL
      INTEGER COUNT1
      INTEGER COUNT2
      INTEGER COMM

      COMMON /ERRCOM/PASS, FAIL, COUNT1, COUNT2, COMM

      INTEGER CLASS

      CHARACTER*(IOSIZE) INFOBUF

!
!   Handles errors for Dup'd Comm
!
      IF (COUNT2 .EQ. 0) THEN
!         Bad rank on MPI_SEND
        CALL MPI_ERROR_CLASS(CODE, CLASS, IERR)
        IF (CLASS .NE. MPI_ERR_RANK) THEN
          INFOBUF='Third error returned incorrect code'
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        ELSE
         PASS = PASS + 1
        END IF
      ELSE
        INFOBUF='Errhandler 2 called too many times'
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        FAIL = FAIL + 1
      END IF

      IF (COMMIN .NE. COMM) THEN
        INFOBUF='Errhandler 1 called w/ incorrect communicator'
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        FAIL = FAIL + 1
      END IF

      COUNT2 = COUNT2 + 1
      ERR2 = 0

      END


      PROGRAM MAIN
        USE MPI

      INCLUDE 'mpitestf90.h'
      INCLUDE 'externalf90.h'


      EXTERNAL ERR1
      EXTERNAL ERR2

!                           counts total number of PASSes/FAILures/VERIFY
      INTEGER PASS
      INTEGER FAIL              
!                           counts number of times errhandler is called
      INTEGER COUNT1
      INTEGER COUNT2

      INTEGER COMM

      COMMON /ERRCOM/PASS, FAIL, COUNT1, COUNT2, COMM

      INTEGER ERRH1
      INTEGER ERRH2
!                           return value from MPI calls
      INTEGER IERR
      INTEGER ERR
      LOGICAL FLAG
!                           error class of IERR
      INTEGER ERRORCLASS
      INTEGER ERRSIZE 

      CHARACTER*(IOSIZE)  INFOBUF
      CHARACTER*32   TESTNAME
      CHARACTER*(MPI_MAX_ERROR_STRING) ERRORSTRING


 98   FORMAT(A,INT_FMT,A)
 99   FORMAT(A,INT_FMT)
 100  FORMAT(A200)


!
!  Initialize the MPI environment and test environment.
!
      CALL MPI_INIT(IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF,99) 'MPI_INIT() returned', IERR 
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF

      PASS = 0
      FAIL = 0

      TESTNAME='MPI_Errhandler_free'
      CALL MPITEST_INIT(IERR)
      IF (MPITEST_ME .EQ. 0) THEN
        INFOBUF = 'Starting test '//TESTNAME
        CALL MPITEST_MESSAGE(MPITEST_INFO0, INFOBUF)
      END IF

      CALL MPI_COMM_DUP(MPI_COMM_WORLD, COMM, IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF, 99) &
           'MPI_COMM_DUP() non-zero return code', IERR
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        INFOBUF = ' '
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, ERRSIZE, ERR)
        CALL MPITEST_MESSAGE(MPITEST_FATAL, ERRORSTRING)
      END IF

!
!    Create 2 errorhandlers
!

      CALL MPI_ERRHANDLER_CREATE(ERR1, ERRH1, IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF, 99) &
           'MPI_ERRHANDLER_CREATE() #1 non-zero return code', IERR
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        INFOBUF = ' '
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, ERRSIZE, ERR)
        CALL MPITEST_MESSAGE(MPITEST_FATAL, ERRORSTRING)
      END IF
      PASS = PASS + 1


      CALL MPI_ERRHANDLER_CREATE(ERR2, ERRH2, IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF, 99) &
           'MPI_ERRHANDLER_CREATE() #2 non-zero return code', IERR
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        INFOBUF = ' '
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, ERRSIZE, ERR)
        CALL MPITEST_MESSAGE(MPITEST_FATAL, ERRORSTRING)
      END IF
      PASS = PASS + 1

!
!    Set 2 errorhandlers
!

      CALL MPI_ERRHANDLER_SET(MPI_COMM_WORLD, ERRH1, IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF, 99) &
           'MPI_ERRHANDLER_SET() #1 non-zero return code', IERR
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        INFOBUF = ' '
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, ERRSIZE, ERR)
        CALL MPITEST_MESSAGE(MPITEST_FATAL, ERRORSTRING)
      END IF
      PASS = PASS + 1

      CALL MPI_ERRHANDLER_SET(COMM, ERRH2, IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF, 99) &
           'MPI_ERRHANDLER_SET() #2 non-zero return code', IERR
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        INFOBUF = ' '
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, ERRSIZE, ERR)
        CALL MPITEST_MESSAGE(MPITEST_FATAL, ERRORSTRING)
      END IF
      PASS = PASS + 1

!
!   Free the 2 errhandlers
!
      CALL MPI_ERRHANDLER_FREE(ERRH1, IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF, 99) &
           'MPI_ERRHANDLER_FREE() #1 non-zero return code', IERR
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        INFOBUF = ' '
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, ERRSIZE, ERR)
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, ERRORSTRING)
      ELSE IF (ERRH1 .NE. MPI_ERRHANDLER_NULL) THEN
        INFOBUF=  &
           'Errhandler #1 .NE. MPI_ERRHANDLER_NULL after FREE call'
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
      ELSE
        PASS = PASS + 1
      END IF


      CALL MPI_ERRHANDLER_FREE(ERRH2, IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF, 99) &
           'MPI_ERRHANDLER_FREE() #2 non-zero return code', IERR
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        INFOBUF = ' '
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, ERRSIZE, ERR)
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, ERRORSTRING)
      ELSE IF (ERRH2 .NE. MPI_ERRHANDLER_NULL) THEN
        INFOBUF=  &
           'Errhandler #2 .NE. MPI_ERRHANDLER_NULL after FREE call'
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
      ELSE
        PASS = PASS + 1
      END IF

!
!   Now create some errors and verify that they work
!

!      MPI_COMM_WORLD
      CALL MPI_SEND(IERR, -1, MPI_INTEGER, 0, 0, MPI_COMM_WORLD, IERR)

!      This should also be handled by MPI_COMM_WORLD
      CALL MPI_BARRIER(MPI_COMM_NULL, IERR)

!      Dup'ed comm
      CALL MPI_SEND(IERR, 1, MPI_INTEGER, MPITEST_NUMP+1, 0, COMM, IERR)

      IF (COUNT1 .NE. 2) THEN
        INFOBUF = 'Errhandler 1 was not called correctly on error'
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        FAIL = FAIL + 1
      END IF

      IF (COUNT2 .NE. 1) THEN
        INFOBUF = 'Errhandler 2 was not called correctly on error'
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        FAIL = FAIL + 1
      END IF

!
!  Report overall results
!
      CALL MPITEST_REPORT(PASS, FAIL, 0, TESTNAME)

      CALL MPI_FINALIZE(IERR)

      END
