! ----------------------------------------------------------------------
! MESSAGE PASSING INTERFACE TEST CASE SUITE
! 
! Copyright - 1996 Intel Corporation
! 
! Intel Corporation hereby grants a non-exclusive license under Intel's
! copyright to copy, modify and distribute this software for any purpose 
! and without fee, provided that the above copyright notice and the
! following paragraphs appear on all copies.
! 
! Intel Corporation makes no representation that the test cases
! comprising this suite are correct or are an accurate representation
! of any standard.
! 
! IN NO EVENT SHALL INTEL HAVE ANY LIABILITY FOR ANY DIRECT, INDIRECT
! OR SPECULATIVE DAMAGES, (INCLUDING WITHOUT LIMITING THE FOREGOING,
! CONSEQUENTIAL, INCIDENTAL AND SPECIAL DAMAGES) INCLUDING, BUT NOT
! LIMITED TO INFRINGEMENT, LOSS OF USE, BUSINESS INTERRUPTIONS, AND
! LOSS OF PROFITS, IRRESPECTIVE OF WHETHER INTEL HAS ADVANCE NOTICE OF
! THE POSSIBILITY OF ANY SUCH DAMAGES.
! 
! INTEL CORPORATION SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING,
! BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
! FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  THE SOFTWARE PROVIDED
! HEREUNDER IS ON AN "AS IS" BASIS AND INTEL CORPORATION HAS NO
! OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS OR
! MODIFICATIONS.
! 
! ----------------------------------------------------------------------
!***********************************************************************
!  
!                     MPI_Intercomm_create_err9
! 
! This test verifies that the correct error is returned if 
! MPI_Intercomm_create() is called with invalid arguments.
! 
! MPI_Intercomm_create error tests
! -----------------------------------
!  1)  local_comm=MPI_COMM_NULL ............................ [COMM]
!  2)  peer_comm=MPI_COMM_NULL ............................. [COMM]
!  3)  removed (invalid test)
!  4)  removed (invalid test)
!  5)  peer_comm not containing local and remote leaders ... [ARG/COMM/OTHER]
!  6)  removed
!  7)  local_leader=MPI_ANY_SOURCE ......................... [ARG/OTHER/RANK]
!  8)  remote_leader=MPI_ANY_SOURCE ........................ [ARG/OTHER/RANK]
!  9)  local_leader not a valid rank in local_comm ......... [RANK]
! 10) remote_leader not a valid rank in peer_comm .........  [RANK]
! 
! In all cases, expect to receive appropriate error.
! 
! MPI Calls dependencies for this test:
!   MPI_Intercomm_create(), MPI_Init(), MPI_Finalize()
!   MPI_Error_string(), MPI_Comm_split(), MPI_Comm_dup(),
!  
!***********************************************************************

#include "f90options.h"

      PROGRAM MAIN
        USE MPI
!
      INCLUDE 'mpitestf90.h'
      INCLUDE 'externalf90.h'

!
      INTEGER    MIN_NODES
      PARAMETER (MIN_NODES = 2)
!                                  Minimum nodes needed to run job
      CHARACTER  TEST_DESC*(*)
      PARAMETER( TEST_DESC = &
        'MPI_Intercomm_create: call with invalid LOCAL_LEADER')
!
      CHARACTER  ERR_DESC*(*)
      PARAMETER (ERR_DESC  = 'Invalid LOCAL_LEADER')
!
      INTEGER PASS
      INTEGER FAIL              
!                           counts total number of PASSes/FAILures
      INTEGER IERR
      INTEGER ERR
!                           return value from MPI calls
      INTEGER ERRORCLASS
      INTEGER SIZE 
!                           error class of IERR
      INTEGER RESULT 
!                           result from intercomm_create
      CHARACTER*(IOSIZE)  INFOBUF
!                           buffer for passing messages to MPITEST
      CHARACTER*(MPI_MAX_ERROR_STRING)  ERRORSTRING
!                           string describing error flag
      CHARACTER*32   TESTNAME
!                           name of this test
      INTEGER LOCAL_COMM
!                           local intra-communicator
      INTEGER PEER_COMM
!                           "peer" inter-communicator
      INTEGER NEWINTERCOMM
!                           the new inter-communicator
      INTEGER LOCAL_LEADER  
!                           rank of the local group leader in local_comm
      INTEGER REMOTE_LEADER
!                           rank of the remote leader in peer_comm
      INTEGER TAG
!                           "safe" tag
      INTEGER COLOR
!                           discriminator used by MPI_Comm_split 
      INTEGER KEY
!                           controls ordering in new LOCAL_COMMs
!-----------------------------------------------------------------------------
  98  FORMAT(A,A)
 100  FORMAT(A200)
 101  FORMAT(A, A, INT_FMT)
 102  FORMAT(A, INT_FMT, A, INT_FMT)
 103  FORMAT(A, INT_FMT)
 104  FORMAT(A, INT_FMT, A)
!
!  Initialize the MPI environment and test environment.
!
      CALL MPI_INIT(IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF,103) 'MPI_INIT() returned', IERR 
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
      TESTNAME='MPI_Intercomm_create_err9'
      CALL MPITEST_INIT(IERR)
      IF (MPITEST_ME .EQ. 0) THEN
        INFOBUF = 'Starting test '//TESTNAME
        CALL MPITEST_MESSAGE(MPITEST_INFO0, INFOBUF)
      END IF

      PASS = 0
      FAIL = 0
!
!  Test if we have enough nodes to run test
!
      IF (MPITEST_NUMP .LT.  MIN_NODES)  THEN 
        FAIL = FAIL + 1
        WRITE(INFOBUF, 102)  &
          'Not enought nodes to run test,  #nodes = ', MPITEST_NUMP, &
          '  Need at least ', MIN_NODES
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
! 
!  Set an errorhandler so we get control back.  
! 
      CALL MPI_ERRHANDLER_SET(MPI_COMM_WORLD, MPI_ERRORS_RETURN, IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        WRITE(INFOBUF, 103)  &
          ' MPI_ERRORHANDLER_SET(MPI_COMM_WORLD)  returned ', IERR
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
        WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
!
!     Set variables for call
!
!     set colors -- the lower half are 0 and the upper half are 1
       COLOR = 0
       IF ( MPITEST_ME .GE. MPITEST_NUMP/2 ) COLOR = 1
!     set keys -- the first and last processes are 0 the rest are one */
       KEY = 1
       IF (MPITEST_ME .EQ. 0   .OR.   &
           MPITEST_ME .EQ.  (MPITEST_NUMP - 1) ) KEY = 0
!
!-----------------------  MPI_Comm_split  -----------------------------
!
!     Use a call to MPI_Comm_split() to create the local_comm values */
!
      CALL MPI_COMM_SPLIT(MPI_COMM_WORLD, COLOR, KEY, LOCAL_COMM, IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        WRITE(INFOBUF, 103) ' MPI_COMM_SPLIT returned ', IERR
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
        WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
!
!     Set LOCAL_LEADER -- all are 0 due to the use of keys above
!
      LOCAL_LEADER = 0
!
!     PEER_COMM is a copy of MPI_COMM_WORLD
!
!-----------------------  MPI_Comm_dup  -----------------------------
!
      CALL MPI_COMM_DUP(MPI_COMM_WORLD, PEER_COMM, IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        WRITE(INFOBUF, 103) ' MPI_COMM_DUP returned ', IERR
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
        WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
!
!     Set remote_leader -- first and last ranks in MPI_COMM_WORLD */
!
      REMOTE_LEADER = 0
      IF(MPITEST_ME .LT. MPITEST_NUMP/2)  &
                                       REMOTE_LEADER = MPITEST_NUMP-1
!
!     choose a tag -- anything will do since we only need one
!
      TAG = 42
!
!     Setup errorhandlers on local_comm so the tests work 
!
      CALL MPI_ERRHANDLER_SET(LOCAL_COMM, MPI_ERRORS_RETURN, IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        WRITE(INFOBUF, 103)  &
          ' MPI_ERRORHANDLER_SET(LOCAL_COMM) returned ', IERR
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
        WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
!
!     Change some of the values to generate the desired errors
!
      LOCAL_LEADER = MPITEST_NUMP
!
!     Calling MPI_Intercomm_create with bad arguments 
!
      IF(MPITEST_ME .EQ. 0) THEN
        WRITE(INFOBUF, 98)  &
        'Calling MPI_Intercomm_create with ', ERR_DESC
        CALL MPITEST_MESSAGE(MPITEST_INFO1, INFOBUF)
        END IF
!
!-----------------------  MPI_Intercomm_create  -----------------------------
!
        CALL MPI_INTERCOMM_CREATE(LOCAL_COMM, LOCAL_LEADER, PEER_COMM,  &
                                REMOTE_LEADER, TAG, NEWINTERCOMM, IERR)
        IF (IERR .EQ. MPI_SUCCESS) THEN
          FAIL = FAIL + 1
          WRITE(INFOBUF, 98) TEST_DESC, ' did not FAIL'
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        ELSE
          CALL MPI_ERROR_CLASS(IERR, ERRORCLASS, ERR)
          IF (ERRORCLASS .NE. MPI_ERR_RANK) THEN
            FAIL = FAIL + 1
            WRITE(INFOBUF,101) TEST_DESC, ' Returned',  ERRORCLASS 
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
            WRITE(INFOBUF,104)  &
              'Expected MPI_ERR_RANK(',   MPI_ERR_RANK, ')'
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
            ERRORSTRING = ' '
            CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
            WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ELSE
            PASS = PASS + 1
            WRITE(INFOBUF,102) 'IERR = ', IERR, '  ERRORCLASS = ', &
              ERRORCLASS
            CALL MPITEST_MESSAGE(MPITEST_INFO2, INFOBUF)
            ERRORSTRING = ' '
            CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
            WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
            CALL MPITEST_MESSAGE(MPITEST_INFO1, INFOBUF)
          END IF
        END IF
      CALL MPITEST_REPORT(PASS, FAIL, 0, TESTNAME)
      CALL MPI_FINALIZE(IERR)
      END
