! ----------------------------------------------------------------------
! MESSAGE PASSING INTERFACE TEST CASE SUITE
! 
! Copyright - 1996 Intel Corporation
! 
! Intel Corporation hereby grants a non-exclusive license under Intel's
! copyright to copy, modify and distribute this software for any purpose 
! and without fee, provided that the above copyright notice and the
! following paragraphs appear on all copies.
! 
! Intel Corporation makes no representation that the test cases
! comprising this suite are correct or are an accurate representation
! of any standard.
! 
! IN NO EVENT SHALL INTEL HAVE ANY LIABILITY FOR ANY DIRECT, INDIRECT
! OR SPECULATIVE DAMAGES, (INCLUDING WITHOUT LIMITING THE FOREGOING,
! CONSEQUENTIAL, INCIDENTAL AND SPECIAL DAMAGES) INCLUDING, BUT NOT
! LIMITED TO INFRINGEMENT, LOSS OF USE, BUSINESS INTERRUPTIONS, AND
! LOSS OF PROFITS, IRRESPECTIVE OF WHETHER INTEL HAS ADVANCE NOTICE OF
! THE POSSIBILITY OF ANY SUCH DAMAGES.
! 
! INTEL CORPORATION SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING,
! BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
! FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  THE SOFTWARE PROVIDED
! HEREUNDER IS ON AN "AS IS" BASIS AND INTEL CORPORATION HAS NO
! OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS OR
! MODIFICATIONS.
! 
! ----------------------------------------------------------------------
!***********************************************************************
!  		  Test for MPI keyvals and attributes.
!  
!  This test creates MPI Keyvals (create and free), and
!  assigns attributes (get, put, delete) to them on various communicator 
!  types as provided bt the MPITEST environment.
!  
!  It tests: 1) User copy function, flag=1 / user delete function
!            2) MPI_NULL_COPY_FN           / user delete function
!            3) MPI_DUP_FN                 / user delete function
!            4) User copy function, flag=0 / MPI_NULL_DELETE_FN
!  on the Keyval_create call, calling Attr_get, Attr_put and Attr_free on
!  each communicator and a dup'ed communicator and checking for the appropriate
!  results of each.  The extra_state is incremented in each user-defined
!  callback function, to validate that it was indeed called.
!  
!  
!  Test history:
!     1  08/28/96     gt       Original version
!  
! NOTE:  9/96
!    The MPI Standard(6/95), Section 5.7.1(Page 169)  gives a prototype
!    for a Copy callback function (SUBROUTINE COPY_FUNCTION), and
!    a Callback deletion function (SUBROUTINE DELETE_FUNCTION).
!    The last formal parameter in both of these subroutines is an
!    INTEGER error variable, IERR.  This variable is not in the INTEL MPI
!    code, and when run with the IERR parameter included in the declaration,
!    the program crashes.  This program references Callback-function
!    calls to subroutines at the end of this file 
!***********************************************************************

#include "f90options.h"

      INCLUDE 'mpitest_cfgf90.h'
!
      PROGRAM MAIN
        USE MPI
!
      INCLUDE 'mpitestf90.h'
      INCLUDE 'externalf90.h'

!
!  Program constants (Parameters)
!
!
      CHARACTER*32          TESTNAME
      PARAMETER( TESTNAME = 'MPI_KEYVAL1')
!
!  General variables that occur in all programs
!
      INTEGER    CNUMBR
!                               index for communications loop
      INTEGER    COMM_INDEX
!                               the array index of the current comm
      INTEGER    COMM_TYPE
!                               the index of the current communicator type
      INTEGER    COMM_COUNT
!                               loop counter for communicator loop
      INTEGER    FTEST_NUMP
!                               number of ranks in the current communicator
      INTEGER    LOOP_COUNT
!                               counts total number of failures, loops
      INTEGER    COMM
!                               MPITEST  communicator under test
      INTEGER    IERR
      INTEGER    ERR
!                               return value from MPI calls
      INTEGER    ERRORCLASS
      INTEGER    SIZE 
!                               error class of IERR
      INTEGER    ERRSIZE
!                               length of error message
      INTEGER    PASS
      INTEGER    FAIL              
      INTEGER    VERIFY
!                           counts total number of PASSes/FAILures/Verifies
      CHARACTER*(IOSIZE)    INFOBUF
!                           buffer for passing messages to MPITEST
      CHARACTER*(MPI_MAX_ERROR_STRING)  ERRORSTRING
!                           string describing error flag
!
!  Variables specific to this program
!
      LOGICAL    INTERCOMM
!                               intercommunicator flag
      LOGICAL    FOUND
!                               FLAG from MPI_Attr_get
      INTEGER    RESULT
!                               output from MPI_Comm_compare
      INTEGER    GROUP
!                               group handle
      INTEGER    COMM2
!                               for split communicator
      INTEGER    KEY
      INTEGER    ATTR
!                               MPI_Key
      INTEGER    COMM_SIZE
!                               communicator size
      INTEGER    EXTRA1
      INTEGER    EXTRA2
      INTEGER    EXTRA3
      INTEGER    EXTRA4
!                               for keyval creation/use
      INTEGER    KEY1, VALUE1
      INTEGER    KEY2, VALUE2
      INTEGER    KEY3, VALUE3
      INTEGER    KEY4, VALUE4


      EXTERNAL   COPY_FUNCTION1, COPY_FUNCTION2
      EXTERNAL   DELETE_FUNCTION
!                               for Keval_create call
      

!-----------------------------------------------------------------------------
 98   FORMAT(A,A )
 99   FORMAT(A,INT_FMT)
 100  FORMAT(A200)
 101  FORMAT(A, A, INT_FMT, A, INT_FMT)
 102  FORMAT(A,INT_FMT, A,INT_FMT, A,INT_FMT, A,INT_FMT, A,INT_FMT , &
             A,INT_FMT, A,INT_FMT, A,INT_FMT, A,INT_FMT, A,INT_FMT)
 103  FORMAT(A, INT_FMT, A, A, A, INT_FMT)
 104  FORMAT(A, A, A, INT_FMT, A)
 105  FORMAT(A, INT_FMT, A, A, INT_FMT, A)
!===========================================================================
!
!  Initialize the MPI environment
!
      CALL MPI_INIT(IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF,99) 'MPI_INIT() returned', IERR
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
!
!  Initialize the MPITEST environment
!
      CALL MPITEST_INIT(IERR)
!
      IF (MPITEST_ME .EQ. 0) THEN
        INFOBUF = 'Starting test '//TESTNAME
        CALL MPITEST_MESSAGE(MPITEST_INFO0, INFOBUF)
      END IF
!
!  Get number of communicators
!
      CALL MPITEST_NUM_COMM_SIZES(COMM_COUNT)

      FAIL = 0
      LOOP_COUNT = 0
      VERIFY = 0
!
!---------------------  Loop over Communicators  ------------------------
!
      DO 4000 CNUMBR = 1,COMM_COUNT
        CALL MPITEST_GET_COMM_INDEX(CNUMBR, COMM_INDEX)
        CALL MPITEST_GET_COMM_TYPE(CNUMBR, COMM_TYPE)
        CALL MPITEST_GET_COMMUNICATOR(COMM_TYPE, COMM_INDEX, COMM, &
              COMM_SIZE)
!
!       Skip to end of loop if this node is not part of current
!       communicator
!
        IF (MPITEST_CURRENT_RANK .EQ. MPI_UNDEFINED ) GO TO 3800

        EXTRA1 = 0
        EXTRA2 = 0
        EXTRA3 = 0
        EXTRA4 = 0
!
!-------------------------  Create Keyval #1  ----------------------------
!

        CALL   MPI_KEYVAL_CREATE(COPY_FUNCTION1, &
                                 DELETE_FUNCTION, &
                                 KEY1, &
                                 EXTRA1, &
                                 IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          FAIL = FAIL + 1
          WRITE(INFOBUF,103) 'Non-Zero return code (', IERR,  &
           ')  From:  MPI_KEYVAL_CREATE #1'
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
        END IF
!
!     Keyval exists, but no attribute set yet, so expect found = 0
!
        CALL MPI_ATTR_GET( COMM, &
                           KEY1, &
                           ATTR, &
                           FOUND, &
                           IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,105) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_Attr_get(KEY1) before put ', &
            ' (', 'COMM_INDEX', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        ELSE IF  (FOUND .NEQV.  .FALSE.) THEN
          WRITE(INFOBUF,104)  &
            'MPI_Attr_get(KEY1) reported KEY1 found before ',  &
            'initialise   (', 'COMM_INDEX', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!  Set a value for the attribute
!
        VALUE1 = MPITEST_ME
!
        CALL MPI_ATTR_PUT(COMM,  &
                          KEY1,  &
                          VALUE1,  &
                          IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          FAIL = FAIL + 1
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_ATTR_PUT(KEY1)   ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        END IF
!
!     Get the value of the attribute
!
        CALL MPI_ATTR_GET( COMM, &
                           KEY1, &
                           ATTR, &
                           FOUND, &
                           IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,105) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_Attr_get(KEY1) AFTER put ', &
            ' (', 'COMM_INDEX', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        ELSE IF  (FOUND .NEQV.  .TRUE.) THEN
          WRITE(INFOBUF,104)  &
            'MPI_Attr_get(KEY1) reported KEY1 NOT found after ',  &
            'initialised   (', 'COMM_INDEX', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        ELSE IF  (ATTR .NE. VALUE1) THEN
          WRITE(INFOBUF,104)  &
            'MPI_Attr_get(KEY1) VALUE =  ',  ATTR, &
            '  Expected =', VALUE1, &
            '  (', 'COMM_INDEX', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!     Dup the Communicator, attribute should follow
!
        CALL MPI_COMM_DUP(COMM,  &
                          COMM2,  &
                          IERR)
!
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_COMM_DUP ( COMM_INDEX =  ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!  Get the value for the attribute
!
        CALL MPI_ATTR_GET( COMM2, &
                           KEY1, &
                           ATTR, &
                           FOUND, &
                           IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_Attr_get(KEY1) on dup''d communicator '
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        ELSE IF  (FOUND .NEQV.  .TRUE.) THEN
          WRITE(INFOBUF,104)  &
            'MPI_Attr_get reported KEY1, but not found on dup''d ',  &
            'communicator   (', 'COMM_INDEX', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!  Delete the Attribute on COMM 
!
        CALL MPI_ATTR_DELETE(COMM, &
                             KEY1,  &
                             IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_ATTR_DELETE (1)(KEY1)  ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
!
!     Keyval exists, but no longer is  attribute set yet, so expect found = 0
!
        CALL MPI_ATTR_GET( COMM, &
                           KEY1, &
                           ATTR, &
                           FOUND, &
                           IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,105) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_Attr_get(KEY1) after deleted ', &
            ' (', 'COMM_INDEX', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        ELSE IF  (FOUND .NEQV.  .FALSE.) THEN
          WRITE(INFOBUF,104)  &
            'MPI_Attr_get(KEY1) reported KEY1 found after deleted',  &
            '   (', 'COMM_INDEX', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!  Get the value for the attribute on the dup'ed comm, should still be there
!
        CALL MPI_ATTR_GET( COMM2, &
                           KEY1, &
                           ATTR, &
                           FOUND, &
                           IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_Attr_get(KEY1) on dup''ed communicator '
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        ELSE IF  (FOUND .NEQV.  .TRUE.) THEN
          WRITE(INFOBUF,104)  &
            'MPI_Attr_get #2 reported key not found on dup''ed ', &
            'communicator   (', 'COMM_INDEX', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        ELSE IF  (ATTR .NE. VALUE1) THEN
          WRITE(INFOBUF,102)  &
            'MPI_Attr_get(KEY1) #2 on dup''ed comm, ATTR =',   &
            ATTR, '  Expected = ', VALUE1,  &
            '  ( COMM_INDEX =', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!  Free the dup'ed  Communicator
!
        CALL MPI_COMM_FREE(COMM2, IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_COMM_FREE(KEY1)  ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1

!-----------------------------------------------------------------------------
!                             KEYVAL2
!-----------------------------------------------------------------------------
!
!-------------------------  Create Keyval #2  ----------------------------
!

        CALL   MPI_KEYVAL_CREATE(MPI_NULL_COPY_FN, &
                                 DELETE_FUNCTION, &
                                 KEY2, &
                                 EXTRA2, &
                                 IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          FAIL = FAIL + 1
          WRITE(INFOBUF,103) 'Non-Zero return code (', IERR,  &
           ')  From:  MPI_KEYVAL_CREATE #1'
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
        END IF
!
!     Keyval exists, but no attribute set yet, so expect found = 0
!
        CALL MPI_ATTR_GET( COMM, &
                           KEY2, &
                           ATTR, &
                           FOUND, &
                           IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,105) 'Non-Zero return code (', IERR,  &
                ')  From:  MPI_Attr_get(KEY2) before put ', &
              ' (', 'COMM_INDEX', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        ELSE IF  (FOUND .NEQV.  .FALSE.) THEN
          WRITE(INFOBUF,104)  &
            'MPI_Attr_get(KEY2) reported KEY2 found before ',  &
            'initialise   (', 'COMM_INDEX', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!  Set a value for the attribute
!
        VALUE2 = MPITEST_ME
!
        CALL MPI_ATTR_PUT(COMM,  &
                          KEY2,  &
                          VALUE2,  &
                          IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          FAIL = FAIL + 1
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_ATTR_PUT(KEY2)   ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        END IF
!
!     Get the value of the attribute
!
        CALL MPI_ATTR_GET( COMM, &
                           KEY2, &
                           ATTR, &
                           FOUND, &
                           IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,105) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_Attr_get(KEY2) AFTER put ', &
            ' (', 'COMM_INDEX', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        ELSE IF  (FOUND .NEQV.  .TRUE.) THEN
          WRITE(INFOBUF,104)  &
            'MPI_Attr_get(KEY2) reported KEY2 NOT found after ',  &
            'initialised   (', 'COMM_INDEX', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        ELSE IF  (ATTR .NE. VALUE2) THEN
          WRITE(INFOBUF,104)  &
            'MPI_Attr_get(KEY2) VALUE =  ',  ATTR, &
            '  Expected =', VALUE2, &
            '  (', 'COMM_INDEX', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!     Dup the Communicator, attribute should NOT follow
!
        CALL MPI_COMM_DUP(COMM,  &
                          COMM2,  &
                          IERR)
!
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_COMM_DUP (KEY2) (COMM_INDEX =',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!  Get the value for the attribute, it should NOT be found
!     
        CALL MPI_ATTR_GET( COMM2, &
                           KEY2, &
                           ATTR, &
                           FOUND, &
                           IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_Attr_get(KEY2) on dup''d communicator '
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        ELSE IF  (FOUND .NEQV.  .FALSE.) THEN
          WRITE(INFOBUF,101)  &
            'MPI_Attr_get reported KEY2 found on dup''ed  ',  &
            'communicator   (COMM_INDEX =', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!  Delete the Attribute on COMM 
!
        CALL MPI_ATTR_DELETE(COMM, &
                             KEY2,  &
                             IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_ATTR_DELETE (2)  ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
!
!     Keyval exists, but no longer is  attribute set yet, so expect found = 0
!
        CALL MPI_ATTR_GET( COMM, &
                           KEY2, &
                           ATTR, &
                           FOUND, &
                           IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,105) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_Attr_get(KEY2) after deleted ', &
              ' (', 'COMM_INDEX', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        ELSE IF  (FOUND .NEQV.  .FALSE.) THEN
          WRITE(INFOBUF,104)  &
            'MPI_Attr_get(KEY2) reported KEY2 found after deleted',  &
            '   (', 'COMM_INDEX', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!     Get the value for the attribute on the dup'ed comm, 
!     should still NOT be there
!
        CALL MPI_ATTR_GET( COMM2, &
                           KEY2, &
                           ATTR, &
                           FOUND, &
                           IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_Attr_get(KEY2) #2 on dup''ed communicator'
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        ELSE IF  (FOUND .NEQV.  .FALSE.) THEN
          WRITE(INFOBUF,101)  &
            'MPI_Attr_get #2 reported KEY2  found on dup''ed ', &
            'communicator   (COMM_INDEX =', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!  Free the dup'ed  Communicator
!
        CALL MPI_COMM_FREE(COMM2, IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_COMM_FREE(KEY2)  ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!-----------------------------------------------------------------------------
!                           KEYVAL3
!-----------------------------------------------------------------------------
!
!-------------------------  Create Keyval #3  ----------------------------
!
        CALL   MPI_KEYVAL_CREATE(MPI_DUP_FN, &
                                 DELETE_FUNCTION, &
                                 KEY3, &
                                 EXTRA3, &
                                 IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          FAIL = FAIL + 1
          WRITE(INFOBUF,103) 'Non-Zero return code (', IERR,  &
           ')  From:  MPI_KEYVAL_CREATE #1'
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
        END IF
!
!     Keyval exists, but no attribute set yet, so expect found = 0
!
        CALL MPI_ATTR_GET( COMM, &
                           KEY3, &
                           ATTR, &
                           FOUND, &
                           IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,105) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_Attr_get(KEY3) before put ', &
            ' (', 'COMM_INDEX', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        ELSE IF  (FOUND .NEQV.  .FALSE.) THEN
          WRITE(INFOBUF,104)  &
            'MPI_Attr_get(KEY3) reported KEY3 found before ',  &
            'initialised  (', 'COMM_INDEX', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!  Set a value for the attribute
!
        VALUE3 = MPITEST_ME
!
        CALL MPI_ATTR_PUT(COMM,  &
                          KEY3,  &
                          VALUE3,  &
                          IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          FAIL = FAIL + 1
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_ATTR_PUT(KEY3)   ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        END IF
!
!     Get the value of the attribute
!
        CALL MPI_ATTR_GET( COMM, &
                           KEY3, &
                           ATTR, &
                           FOUND, &
                           IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,105) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_Attr_get(KEY3) AFTER put ', &
            ' (', 'COMM_INDEX', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        ELSE IF  (FOUND .NEQV.  .TRUE.) THEN
          WRITE(INFOBUF,104)  &
            'MPI_Attr_get(KEY3) reported KEY3 NOT found after ',  &
            'initialised   (', 'COMM_INDEX', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        ELSE IF  (ATTR .NE. VALUE3) THEN
          WRITE(INFOBUF,104)  &
            'MPI_Attr_get(KEY3) VALUE =  ',  ATTR, &
            '  Expected =', VALUE3, &
            '  (', 'COMM_INDEX', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!     Dup the Communicator, attribute should follow
!
        CALL MPI_COMM_DUP(COMM,  &
                          COMM2,  &
                          IERR)
!
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_COMM_DUP (COMM_INDEX = ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!  Get the value for the attribute
!
        CALL MPI_ATTR_GET( COMM2, &
                           KEY3, &
                           ATTR, &
                           FOUND, &
                           IERR)
        IF(IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_Attr_get(KEY3) on dup''ed communicator '
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        ELSE IF(FOUND .NEQV.  .TRUE.) THEN
          WRITE(INFOBUF,104)  &
            'MPI_Attr_get reported KEY3, but not found on dup''ed ',  &
            'communicator   (', 'COMM_INDEX', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        ELSE IF  (ATTR .NE. VALUE3) THEN
          WRITE(INFOBUF,104)  &
            'MPI_Attr_get(KEY3) VALUE =  ',  ATTR, &
            '  Expected =', VALUE3, &
            '  (', 'COMM_INDEX', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!  Delete the Attribute on COMM 
!
        CALL MPI_ATTR_DELETE(COMM, &
                             KEY3,  &
                             IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_ATTR_DELETE(1) (KEY3)  ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
!
!     Keyval exists, but no longer is  attribute set yet, so expect found = 0
!
        CALL MPI_ATTR_GET( COMM, &
                           KEY3, &
                           ATTR, &
                           FOUND, &
                           IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,105) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_Attr_get(KEY3) after deleted ', &
            ' (', 'COMM_INDEX', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        ELSE IF  (FOUND .NEQV.  .FALSE.) THEN
          WRITE(INFOBUF,104)  &
            'MPI_Attr_get(KEY3) reported KEY3 found after deleted',  &
            '   (', 'COMM_INDEX', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!  Get the value for the attribute on the dup'ed comm, should still be there
!
        CALL MPI_ATTR_GET( COMM2, &
                           KEY3, &
                           ATTR, &
                           FOUND, &
                           IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_Attr_get(KEY3) on dup''ed communicator '
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        ELSE IF  (FOUND .NEQV.  .TRUE.) THEN
          WRITE(INFOBUF,104)  &
            'MPI_Attr_get #2 reported key not found on dup''ed ', &
            'communicator   (', 'COMM_INDEX', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        ELSE IF  (ATTR .NE. VALUE3) THEN
          WRITE(INFOBUF,102)  &
            'MPI_Attr_get(KEY3) #2 on dup''ed comm, ATTR =',   &
                ATTR, '  Expected = ', VALUE3,  &
            '   (COMM_INDEX = ', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!  Free the dup'ed  Communicator
!
        CALL MPI_COMM_FREE(COMM2, IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_COMM_FREE(KEY3)  ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
          END IF
          LOOP_COUNT = LOOP_COUNT + 1
!
!---------------------------------------------------------------------------
!                             KEYVAL 4
!---------------------------------------------------------------------------
!
!--------------------------  Create Keyval #4  -----------------------------
!

        CALL   MPI_KEYVAL_CREATE(COPY_FUNCTION2, &
                                 MPI_NULL_DELETE_FN, &
                                 KEY4, &
                                 EXTRA4, &
                                 IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          FAIL = FAIL + 1
          WRITE(INFOBUF,103) 'Non-Zero return code (', IERR,  &
           ')  From:  MPI_KEYVAL_CREATE #4'
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
        END IF
!
!     Keyval exists, but no attribute set yet, so expect found = 0
!
        CALL MPI_ATTR_GET( COMM, &
                           KEY4, &
                           ATTR, &
                           FOUND, &
                           IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,105) 'Non-Zero return code (', IERR,  &
                ')  From:  MPI_Attr_get(KEY4) before put ', &
              ' (', 'COMM_INDEX', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        ELSE IF  (FOUND .NEQV.  .FALSE.) THEN
          WRITE(INFOBUF,104)  &
            'MPI_Attr_get(KEY4) reported KEY4 found before ',  &
            'initialised   (', 'COMM_INDEX', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!  Set a value for the attribute
!
        VALUE4 = MPITEST_ME
!
        CALL MPI_ATTR_PUT(COMM,  &
                          KEY4,  &
                          VALUE4,  &
                          IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          FAIL = FAIL + 1
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_ATTR_PUT(KEY4)   ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        END IF
!
!     Get the value of the attribute
!
        CALL MPI_ATTR_GET( COMM, &
                           KEY4, &
                           ATTR, &
                           FOUND, &
                           IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,105) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_Attr_get(KEY4) AFTER put ', &
            ' (', 'COMM_INDEX', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        ELSE IF  (FOUND .NEQV.  .TRUE.) THEN
          WRITE(INFOBUF,104)  &
            'MPI_Attr_get(KEY4) reported KEY4 NOT found after ',  &
            'initialised   (', 'COMM_INDEX', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        ELSE IF  (ATTR .NE. VALUE4) THEN
          WRITE(INFOBUF,104)  &
            'MPI_Attr_get(KEY4) VALUE =  ',  ATTR, &
            '  Expected =', VALUE4, &
            '  (', 'COMM_INDEX', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!     Dup the Communicator, attribute should NOT follow
!
        CALL MPI_COMM_DUP(COMM,  &
                          COMM2,  &
                          IERR)
!
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_COMM_DUP (KEY4) (COMM_INDEX =',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!  Get the value for the attribute, it should NOT be found
!     
        CALL MPI_ATTR_GET( COMM2, &
                           KEY4, &
                           ATTR, &
                           FOUND, &
                           IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_Attr_get(KEY4) on dup''d communicator '
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        ELSE IF(FOUND .NEQV.  .FALSE.) THEN
          WRITE(INFOBUF,104)  &
            'MPI_Attr_get reported KEY4 found on dup''ed  ',  &
            'communicator   ( COMM_INDEX =', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!  Delete the Attribute on COMM 
!
        CALL MPI_ATTR_DELETE(COMM, &
                             KEY4,  &
                             IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_ATTR_DELETE (2)  ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
!
!     Keyval exists, but no longer is  attribute set yet, so expect found = 0
!
        CALL MPI_ATTR_GET( COMM, &
                           KEY4, &
                           ATTR, &
                           FOUND, &
                           IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,105) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_Attr_get(KEY4) after deleted ', &
              ' (', 'COMM_INDEX', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        ELSE IF  (FOUND .NEQV.  .FALSE.) THEN
          WRITE(INFOBUF,104)  &
            'MPI_Attr_get(KEY4) reported KEY4 found after deleted',  &
            '   (', 'COMM_INDEX', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!     Get the value for the attribute on the dup'ed comm, 
!     should still NOT be there
!
        CALL MPI_ATTR_GET( COMM2, &
                           KEY4, &
                           ATTR, &
                           FOUND, &
                           IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_Attr_get(KEY4) #2 on dup''ed communicator'
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        ELSE IF  (FOUND .NEQV.  .FALSE.) THEN
          WRITE(INFOBUF,101)  &
            'MPI_Attr_get #2 reported KEY4  found on dup''ed ', &
            'communicator   (COMM_INDEX =', COMM_INDEX, ')' 
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!  Free the dup'ed  Communicator
!
        CALL MPI_COMM_FREE(COMM2, IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          WRITE(INFOBUF,102) 'Non-Zero return code (', IERR,  &
            ')  From:  MPI_COMM_FREE(KEY4)  ( COMM_INDEX ',  &
            COMM_INDEX
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          FAIL = FAIL + 1
        END IF
        LOOP_COUNT = LOOP_COUNT + 1
!
!     Free all KEYVALS
!
      CALL   MPI_KEYVAL_FREE(KEY1,  &
                             IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        WRITE(INFOBUF,105) 'Non-Zero return code (', IERR,  &
          ')  From:  MPI_Keyval_free (KEY1)', &
          '  ( COMM_INDEX =',  COMM_INDEX, ')'
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        ERRORSTRING = ' '
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
        WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
      END IF
!
      IF(KEY1 .NE. MPI_KEYVAL_INVALID)  THEN
        WRITE(INFOBUF,101)  &
          'KEY1 not set to MPI_KEYVAL_INVALID by MPI_KEYVAL_FREE',  &
          ',  ( Comm_Index = ',  COMM_INDEX, ')' 
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        FAIL = FAIL + 1
      END IF
      LOOP_COUNT = LOOP_COUNT + 1

      CALL   MPI_KEYVAL_FREE(KEY2,  &
                             IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        WRITE(INFOBUF,105) 'Non-Zero return code (', IERR,  &
          ')  From:  MPI_Keyval_free (KEY2)', &
          '  ( COMM_INDEX =',  COMM_INDEX, ')'
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        ERRORSTRING = ' '
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
        WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
      END IF
!
      IF(KEY2 .NE. MPI_KEYVAL_INVALID)  THEN
        WRITE(INFOBUF,101)  &
          'KEY2 not set to MPI_KEYVAL_INVALID by MPI_KEYVAL_FREE',  &
          ',  ( Comm_Index = ',  COMM_INDEX, ')' 
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        FAIL = FAIL + 1
      END IF
      LOOP_COUNT = LOOP_COUNT + 1

      CALL   MPI_KEYVAL_FREE(KEY3,  &
                             IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        WRITE(INFOBUF,105) 'Non-Zero return code (', IERR,  &
          ')  From:  MPI_Keyval_free (KEY3)', &
          '  ( COMM_INDEX =',  COMM_INDEX, ')'
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        ERRORSTRING = ' '
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
        WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
      END IF
!
      IF(KEY3 .NE. MPI_KEYVAL_INVALID)  THEN
        WRITE(INFOBUF,101)  &
          'KEY3 not set to MPI_KEYVAL_INVALID by MPI_KEYVAL_FREE',  &
          ',  ( Comm_Index = ',  COMM_INDEX, ')' 
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        FAIL = FAIL + 1
      END IF
      LOOP_COUNT = LOOP_COUNT + 1

      CALL   MPI_KEYVAL_FREE(KEY4,  &
                             IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        WRITE(INFOBUF,105) 'Non-Zero return code (', IERR,  &
          ')  From:  MPI_Keyval_free (KEY4)', &
          '  ( COMM_INDEX =',  COMM_INDEX, ')'
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        ERRORSTRING = ' '
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
        WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
      END IF
!
      IF(KEY4 .NE. MPI_KEYVAL_INVALID)  THEN
        WRITE(INFOBUF,101)  &
          'KEY4 not set to MPI_KEYVAL_INVALID by MPI_KEYVAL_FREE',  &
          ',  ( Comm_Index = ',  COMM_INDEX, ')' 
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        FAIL = FAIL + 1
      END IF
      LOOP_COUNT = LOOP_COUNT + 1
!
!     Check the number of tines the callback frunctions were called
!     This is different for each
!
!------------------------------------------------------------------------------
      IF(EXTRA1 .NE. 3) THEN
        WRITE(INFOBUF,105)  &
          'Callback functions called',  EXTRA1, &
          '  times for KEY1, Expected 3',  &
          '   ( Comm_Index = ',  COMM_INDEX, ')' 
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        FAIL = FAIL + 1
      END IF
      LOOP_COUNT = LOOP_COUNT + 1

!------------------------------------------------------------------------------
      IF(EXTRA2 .NE. 1) THEN
        WRITE(INFOBUF,105)  &
          'Callback functions called',  EXTRA2, &
          '  times for KEY2, Expected 1',  &
          '   ( Comm_Index = ',  COMM_INDEX, ')' 
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        FAIL = FAIL + 1
      END IF
      LOOP_COUNT = LOOP_COUNT + 1
!------------------------------------------------------------------------------
      IF(EXTRA3 .NE. 2) THEN
        WRITE(INFOBUF,105)  &
          'Callback functions called',  EXTRA3, &
          '  times for KEY3, Expected 2',  &
          '   ( Comm_Index = ',  COMM_INDEX, ')' 
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        FAIL = FAIL + 1
      END IF
      LOOP_COUNT = LOOP_COUNT + 1
!------------------------------------------------------------------------------
      IF(EXTRA4 .NE. 1) THEN
        WRITE(INFOBUF,105)  &
          'Callback functions called',  EXTRA4, &
          '  times for KEY4, Expected 1',  &
          '   ( Comm_Index = ',  COMM_INDEX, ')' 
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        FAIL = FAIL + 1
      END IF
      LOOP_COUNT = LOOP_COUNT + 1

      CALL MPITEST_FREE_COMMUNICATOR(COMM_TYPE, COMM)
 3800 CONTINUE
!                Node defined in current communicator
 4000 CONTINUE
!
!     Report overall results
!
      CALL MPITEST_REPORT(LOOP_COUNT - FAIL, FAIL, 0, TESTNAME)
      
      CALL MPI_FINALIZE(IERR)

      END
!------------------------------------------------------------------------------
      SUBROUTINE COPY_FUNCTION1(OLDCOMM, &
                                KEYVAL, &
                                EXTRA_STATE, &
                                ATTRIBUTE_VAL_IN, &
                                ATTRIBUTE_VAL_OUT, &
                                FLAG, IERR)
!
      INCLUDE 'mpitestf90.h'
      INCLUDE 'externalf90.h'

!
      INTEGER  OLDCOMM
      INTEGER  KEYVAL
      INTEGER  EXTRA_STATE
      INTEGER  ATTRIBUTE_VAL_IN
      INTEGER  ATTRIBUTE_VAL_OUT
      INTEGER  IERR
!
      LOGICAL  FLAG
!
      ATTRIBUTE_VAL_OUT = ATTRIBUTE_VAL_IN
      FLAG = .TRUE.
      EXTRA_STATE = EXTRA_STATE + 1
      IERR = MPI_SUCCESS
      RETURN
      END
!------------------------------------------------------------------------------
      SUBROUTINE COPY_FUNCTION2(OLDCOMM, &
                                KEYVAL, &
                                EXTRA_STATE, &
                                ATTRIBUTE_VAL_IN, &
                                ATTRIBUTE_VAL_OUT, &
                                FLAG, IERR)
!
      INCLUDE 'mpitestf90.h'
      INCLUDE 'externalf90.h'

!
      INTEGER  OLDCOMM
      INTEGER  KEYVAL
      INTEGER  EXTRA_STATE
      INTEGER  ATTRIBUTE_VAL_IN
      INTEGER  ATTRIBUTE_VAL_OUT
      INTEGER  IERR
!
      LOGICAL  FLAG
!
      FLAG = .FALSE.
      EXTRA_STATE = EXTRA_STATE + 1
      IERR = MPI_SUCCESS
      RETURN
      END
!------------------------------------------------------------------------------
      SUBROUTINE DELETE_FUNCTION(COMM, &
                                 KEYVAL, &
                                 ATTRIBUTE_VAL, &
                                 EXTRA_STATE, IERR)
!
!
      INCLUDE 'mpitestf90.h'
      INCLUDE 'externalf90.h'

!
      INTEGER  COMM
      INTEGER  KEYVAL
      INTEGER  ATTRIBUTE_VAL
      INTEGER  EXTRA_STATE
!
      INTEGER  IERR
!
      EXTRA_STATE = EXTRA_STATE + 1
      IERR = MPI_SUCCESS
      RETURN
      END
