! ----------------------------------------------------------------------
! MESSAGE PASSING INTERFACE TEST CASE SUITE
! 
! Copyright - 1996 Intel Corporation
! 
! Intel Corporation hereby grants a non-exclusive license under Intel's
! copyright to copy, modify and distribute this software for any purpose 
! and without fee, provided that the above copyright notice and the
! following paragraphs appear on all copies.
! 
! Intel Corporation makes no representation that the test cases
! comprising this suite are correct or are an accurate representation
! of any standard.
! 
! IN NO EVENT SHALL INTEL HAVE ANY LIABILITY FOR ANY DIRECT, INDIRECT
! OR SPECULATIVE DAMAGES, (INCLUDING WITHOUT LIMITING THE FOREGOING,
! CONSEQUENTIAL, INCIDENTAL AND SPECIAL DAMAGES) INCLUDING, BUT NOT
! LIMITED TO INFRINGEMENT, LOSS OF USE, BUSINESS INTERRUPTIONS, AND
! LOSS OF PROFITS, IRRESPECTIVE OF WHETHER INTEL HAS ADVANCE NOTICE OF
! THE POSSIBILITY OF ANY SUCH DAMAGES.
! 
! INTEL CORPORATION SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING,
! BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
! FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  THE SOFTWARE PROVIDED
! HEREUNDER IS ON AN "AS IS" BASIS AND INTEL CORPORATION HAS NO
! OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS OR
! MODIFICATIONS.
! 
! ----------------------------------------------------------------------
! **********************************************************************
!
!               MPI_SENDRECV_REPLACE funtional Root to All test
!
! Ref:  MPI Standard, Section 3.10
!
! This test is implemented with the Root-to-all model. 
! The Root Sends, and  the Node Recv calls were both replaced 
! with Sendrecv_replace calls.  Thus, each Root
! is Sending to the other nodes, and Recving from all other nodes. 
! Each non-Root node Recvs a message from the Root, and, concurrently, 
! Sends a message to Root.
!	
! This test may be run in any communicator, with any data type, and with
! any non-negative message length.
!
! The MPITEST environment provides looping over communicator size and
! type, message length, and data type.  The properties of the loops are
! encoded in configuration arrays in the file config.h .  See the
! MPITEST README for further details.
! **********************************************************************
#include "f90options.h"

      INCLUDE 'mpitest_cfgf90.h'

      PROGRAM TEST
        USE MPI

      INCLUDE 'mpitestf90.h'
      INCLUDE 'externalf90.h'


      INTEGER    COMM_INDEX
!                               the array index of the current comm
      INTEGER    COMM_TYPE
!                               the index of the current communicator type
      INTEGER    COMM_COUNT
!                               loop counter for communicator loop
      INTEGER    COMM_SIZE
!                               number of ranks in the current communicator
      INTEGER    TYPE_COUNT
!                               number of data types to test
      INTEGER    TEST_TYPE
!                               index of current data type
      INTEGER    LENGTH_COUNT
!                               number of data lengths to test
      INTEGER    MAX_LENGTH
!                               max message length to test
      INTEGER    MAX_BYTE_LENGTH
!                               max byte length to test
      INTEGER    MAX_BUFF_LENGTH
!                               max length that will fit in buffers
      INTEGER    BYTE_LENGTH
!                               current byte length to test
      INTEGER    LENGTH
!                               current length to test 

      INTEGER    FAIL
      INTEGER    LOOP_CNT
!                               counts total number of failures, loops
      INTEGER    LOOP_FAIL
!                               counts number of failures in 1 loop
      INTEGER    ERROR
!                               number of data errors found in buffer
      INTEGER    IERR
      INTEGER    ERR
!                               return value from MPI calls
      INTEGER    ROOT
!                               the current root
      INTEGER    COMM
!                               MPI communicator under test
      INTEGER    BARRIER_COMM
!                               MPI communicator for BARRIER
      LOGICAL    FOUND
!                               Boolean for MPI_GET_ATTR

      INTEGER    NODE_CNT
      INTEGER    ROOT_CNT
!                               Length of received message from MPI_GET_COUNT
      INTEGER    GRP_LUP
!                               For INTER, index of comm group
      LOGICAL    INTER_FLAG
!                               Is comm under test an INTER communicator
      INTEGER    RECV_SIZE
      INTEGER    LEFT_RECV_SIZE
      INTEGER    RIGHT_RECV_SIZE
!                              For INTER, number of receivers on each side
      INTEGER    SEND_SIZE
      INTEGER    LEFT_SEND_SIZE
      INTEGER    RIGHT_SEND_SIZE
!                              For INTER, number of senders on each side

      INTEGER    NODE_ALL_LENGTH
!                               Message buffer length for node receive
      INTEGER    ROOT_ALL_LENGTH  
!                               Message buffer length for root to receive
      INTEGER    SEND_GROUP
      INTEGER    RECV_GROUP
!                              For INTER, current group under test
      INTEGER    NTIMES
!                              Number of times to go through loop
!                              INTER=2, INTRA=1
      INTEGER    SENDERS
      INTEGER    SEND_TO
      INTEGER    RECEIVERS
!                              Number of senders/receivers
      LOGICAL    PRINT_NODE
!                              Is node to print informational messages

      INTEGER    I, J, K, L, M

      MPITEST_BUF_TYPE NODE_ALL_BUFFER(MAX_BUFF_SIZE+1)

      MPITEST_BUF_TYPE ROOT_ALL_BUFFER(MAX_BUFF_SIZE+1)

      INTEGER    ERRSIZE

      INTEGER    NODE_STAT(MPI_STATUS_SIZE)
      INTEGER    ROOT_STAT(MPI_STATUS_SIZE)

      MPITEST_BUF_TYPE NODE_VALUE
      MPITEST_BUF_TYPE ROOT_VALUE

      CHARACTER*(IOSIZE)  INFOBUF
      CHARACTER*32   TESTNAME
       CHARACTER*(MPI_MAX_ERROR_STRING) ERRORSTRING

 98   FORMAT(A,INT_FMT,A,INT_FMT)
 99   FORMAT(A,INT_FMT)
 179  FORMAT(A,I6,A,I6,A,I6,A,I6,A,I3,A,INT_FMT)
 189  FORMAT(INT_FMT,A,A,I6,A,I6,A,I6,A,I6,A,I3,A,INT_FMT)

!
!     Initialize the MPI environment
!
      CALL MPI_INIT(IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF,99) 'MPI_INIT() returned', IERR
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
!
!     Initialize the MPITEST environment
!
      CALL MPITEST_INIT(IERR)
      TESTNAME = 'MPI_SENDRECV_REPLACE_rtoa'
      IF (MPITEST_ME .EQ. 0) THEN
        INFOBUF = 'Starting test '//TESTNAME
        CALL MPITEST_MESSAGE(MPITEST_INFO0, INFOBUF)
      END IF

!
!     Initialize the loop/error counters
!
      LOOP_CNT = 0
      FAIL = 0
      LOOP_FAIL = 0
!
!     Get number of communicators, data types and message lengths to test
!
      CALL MPITEST_NUM_COMM_SIZES(COMM_COUNT)
      CALL MPITEST_NUM_DATATYPES(TYPE_COUNT)
      CALL MPITEST_NUM_MESSAGE_LENGTHS(LENGTH_COUNT)

!
!     Get max message length for test
!
      CALL MPITEST_GET_MAX_MESSAGE_LENGTH(MAX_BYTE_LENGTH)

!
!---------------------  Loop over Communicators  ------------------------
!
      DO 100 I=1,COMM_COUNT
        CALL MPITEST_GET_COMM_INDEX(I, COMM_INDEX)
        CALL MPITEST_GET_COMM_TYPE(I, COMM_TYPE)

        CALL MPITEST_GET_COMMUNICATOR(COMM_TYPE, COMM_INDEX, COMM, &
              COMM_SIZE)

!       Skip to end of loop if this node is not part of current
!       communicator

        IF (COMM .EQ. MPI_COMM_NULL) GO TO 201

!       Test for INTER-communicator and initialize appropriately if so

        CALL MPI_COMM_TEST_INTER(COMM, INTER_FLAG, IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          INFOBUF=' '
          WRITE(INFOBUF,99) 'MPI_COMM_TEST_INTER() returned', IERR
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          INFOBUF=' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, ERRSIZE, ERR)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, ERRORSTRING)
          FAIL = FAIL + 1
        END IF

        PRINT_NODE = .FALSE.

        IF (INTER_FLAG) THEN
          IF (MPITEST_INTER .EQ. 0) THEN
!           Left side of intercommunicator
            CALL MPI_COMM_SIZE(COMM, SEND_SIZE, IERR)
            CALL MPI_COMM_REMOTE_SIZE(COMM, RECV_SIZE, IERR)

            LEFT_SEND_SIZE = SEND_SIZE
            LEFT_RECV_SIZE = RECV_SIZE
            RIGHT_SEND_SIZE = RECV_SIZE
            RIGHT_RECV_SIZE = SEND_SIZE

            IF (MPITEST_CURRENT_RANK .EQ. 0) PRINT_NODE = .TRUE.

            CALL MPI_INTERCOMM_MERGE(COMM, .FALSE., BARRIER_COMM, IERR)
            IF (IERR .NE. MPI_SUCCESS) THEN
              INFOBUF=' '
              WRITE(INFOBUF,99) 'MPI_INTERCOMM_MERGE() returned', IERR
              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
              INFOBUF=' '
              CALL MPI_ERROR_STRING(IERR, ERRORSTRING, ERRSIZE, ERR)
              CALL MPITEST_MESSAGE(MPITEST_FATAL, ERRORSTRING)
              FAIL = FAIL + 1
            END IF
          END IF

          IF (MPITEST_INTER .EQ. 1) THEN
!           Right side of intercommunicator
            CALL MPI_COMM_SIZE(COMM, SEND_SIZE, IERR)
            CALL MPI_COMM_REMOTE_SIZE(COMM, RECV_SIZE, IERR)

            LEFT_SEND_SIZE = RECV_SIZE
            LEFT_RECV_SIZE = SEND_SIZE
            RIGHT_SEND_SIZE = SEND_SIZE
            RIGHT_RECV_SIZE = RECV_SIZE

            CALL MPI_INTERCOMM_MERGE(COMM, .FALSE., BARRIER_COMM, IERR)
            IF (IERR .NE. MPI_SUCCESS) THEN
              INFOBUF=' '
              WRITE(INFOBUF,99) 'MPI_INTERCOMM_MERGE() returned', IERR
              CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
              INFOBUF=' '
              CALL MPI_ERROR_STRING(IERR, ERRORSTRING, ERRSIZE, ERR)
              CALL MPITEST_MESSAGE(MPITEST_FATAL, ERRORSTRING)
              FAIL = FAIL + 1
            END IF
          END IF

        ELSE
!       Intracommunicator
          IF (MPITEST_CURRENT_RANK .EQ. 0) PRINT_NODE = .TRUE.
          BARRIER_COMM = COMM
        END IF

!
!---------------------  Loop over Data Types  -------------------------
!
        DO 200 J=1,TYPE_COUNT

          CALL MPITEST_GET_DATATYPE(J, TEST_TYPE)
!
!         Convert the number of bytes in the maximum length message
!         into the number of elements of the current type 
          CALL MPITEST_BYTE_TO_ELEMENT(TEST_TYPE, MAX_BYTE_LENGTH, &
                MAX_LENGTH)
          CALL MPITEST_BYTE_TO_ELEMENT(TEST_TYPE, &
                 MAX_BUFF_SIZE*MPITEST_BUF_EXTENT, MAX_BUFF_LENGTH)

          IF (MAX_LENGTH .GE. (MAX_BUFF_LENGTH - 1)) &
               MAX_LENGTH = MAX_BUFF_LENGTH - 1

!
!-----------------------  Loop over Message Lengths  ------------------------
!
          DO 300 K=1,LENGTH_COUNT
            CALL MPITEST_GET_MESSAGE_LENGTH(K, BYTE_LENGTH)
            CALL MPITEST_BYTE_TO_ELEMENT(TEST_TYPE, BYTE_LENGTH, &
                LENGTH)
            IF (LENGTH .GT. MAX_LENGTH) LENGTH = MAX_LENGTH

            NODE_ALL_LENGTH = LENGTH
            ROOT_ALL_LENGTH = LENGTH
!
!           Loop over different processes being the root
!
            IF (INTER_FLAG) THEN
              NTIMES = 2
            ELSE
              NTIMES = 1
            END IF

            DO 350 GRP_LUP=1, NTIMES
!             If intercommunicators, test each side
              SEND_GROUP = GRP_LUP - 1
              RECV_GROUP = 2 - GRP_LUP

              SENDERS = COMM_SIZE
              RECEIVERS = COMM_SIZE - 1
              IF (INTER_FLAG .AND. (GRP_LUP .EQ. 1)) THEN
                RECEIVERS = LEFT_RECV_SIZE
                SENDERS = LEFT_SEND_SIZE
              ELSE IF (INTER_FLAG .AND. (GRP_LUP .EQ. 2)) THEN
                RECEIVERS = RIGHT_RECV_SIZE
                SENDERS = RIGHT_SEND_SIZE
              END IF

              DO 400 ROOT=0,SENDERS-1
!               Print an informational message
                IF (PRINT_NODE) THEN
                  INFOBUF=' '
                  WRITE (INFOBUF,179)  &
                    'commindex', COMM_INDEX, ', commsize', &
                    COMM_SIZE, ', commtype', COMM_TYPE, ', root', ROOT, &
                    ', datatype', TEST_TYPE, ', length',  LENGTH
                  CALL MPITEST_MESSAGE(MPITEST_INFO1, INFOBUF)
                END IF

!
!               All but root on receive side post receives
!
                IF (((INTER_FLAG .EQV. .FALSE.) .AND.  &
                     (MPITEST_CURRENT_RANK .NE. ROOT)) .OR. &
                    ((INTER_FLAG .EQV. .TRUE.) .AND. &
                     (MPITEST_INTER .EQ. RECV_GROUP))) THEN

                  CALL MPITEST_DATATEMPLATE_INIT(NODE_VALUE,  &
                    MPITEST_CURRENT_RANK, TEST_TYPE)
                  CALL MPITEST_INIT_BUFFER(TEST_TYPE,  &
                    NODE_ALL_LENGTH, &
                    NODE_VALUE, NODE_ALL_BUFFER)

                  LOOP_CNT = LOOP_CNT + 1
                  LOOP_FAIL = 0

!=============================================================================
!                        NODE SENDRECV_REPLACE
!-----------------------------------------------------------------------------
                  CALL MPI_SENDRECV_REPLACE( &
                           NODE_ALL_BUFFER, &
                           NODE_ALL_LENGTH, &
                           MPITEST_MPI_DATATYPES(TEST_TYPE),  &
                           ROOT,  &
                           MPITEST_CURRENT_RANK, &
                           ROOT, &
                           ROOT,                          &
                           COMM,  &
                           NODE_STAT, &
                           IERR)

                  IF (IERR .NE. MPI_SUCCESS) THEN
                    INFOBUF=' '
                    WRITE(INFOBUF,99)  &
                      'NODE MPI_SENDRECV_REPLACE() returned', IERR
                    CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                    INFOBUF=' '
                    CALL MPI_ERROR_STRING(IERR, ERRORSTRING, ERRSIZE, &
                         ERR)
                    CALL MPITEST_MESSAGE(MPITEST_NONFATAL, ERRORSTRING)
                    LOOP_FAIL = LOOP_FAIL + 1
                  END IF
!
!                 Check data for errors and overflow
!
                  CALL MPITEST_DATATEMPLATE_INIT(NODE_VALUE, ROOT, &
                    TEST_TYPE)
                  CALL MPITEST_BUFFER_ERRORS(TEST_TYPE,  &
                    NODE_ALL_LENGTH,  &
                    NODE_VALUE, NODE_ALL_BUFFER, ERROR)
                  IF ((ERROR) .NE. 0) THEN
                    INFOBUF=' '
                    WRITE (INFOBUF,189) ERROR, &
                      ' errors in buffer, ', &
                      'commindex',COMM_INDEX,', commsize', COMM_SIZE, &
                      ', commtype', COMM_TYPE, ', root', ROOT, &
                      ', datatype', TEST_TYPE, ', length',   &
                      NODE_ALL_LENGTH
                    CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                    LOOP_FAIL = LOOP_FAIL + 1
                  END IF
                
!
!                 Call MPI_GET_COUNT and compare value with length received
!
                  CALL MPI_GET_COUNT(NODE_STAT,  &
                      MPITEST_MPI_DATATYPES(TEST_TYPE), NODE_CNT, IERR)
                  IF (IERR .NE. MPI_SUCCESS) THEN
                    INFOBUF=' '
                    WRITE(INFOBUF,99) 'MPI_GET_COUNT() returned', IERR
                    CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                    INFOBUF=' '
                    CALL MPI_ERROR_STRING(IERR, ERRORSTRING, ERRSIZE, &
                         ERR)
                    CALL MPITEST_MESSAGE(MPITEST_NONFATAL, ERRORSTRING)
                    LOOP_FAIL = LOOP_FAIL + 1
                  END IF

                  IF (NODE_ALL_LENGTH - NODE_CNT .NE. 0) THEN
                    INFOBUF=' '
                    WRITE(INFOBUF,98)  &
                        'NODE: Unexpected receive length', &
                        NODE_CNT, ', expected', NODE_ALL_LENGTH
                    CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                    INFOBUF=' '
                    LOOP_FAIL = LOOP_FAIL + 1
                  END IF

!
!                 Check source and tag values.
!
                  IF (NODE_STAT(MPI_TAG) .NE. ROOT) THEN
                    INFOBUF=' '
                    WRITE(INFOBUF,98) 'Unexpected tag value', &
                      NODE_STAT(MPI_TAG), ', expected', ROOT
                    CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                    INFOBUF=' '
                    LOOP_FAIL = LOOP_FAIL + 1
                  END IF

                  IF (NODE_STAT(MPI_SOURCE) .NE. ROOT) THEN
                    INFOBUF=' '
                    WRITE(INFOBUF,98) 'Unexpected source value', &
                      NODE_STAT(MPI_SOURCE), ', expected', ROOT
                    CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                    INFOBUF=' '
                    LOOP_FAIL = LOOP_FAIL + 1
                  END IF

                  IF (LOOP_FAIL .NE. 0) FAIL = FAIL + 1
                END IF
!
!                    Root
!

                IF (((INTER_FLAG .EQV. .FALSE.) .AND.  &
                     (MPITEST_CURRENT_RANK .EQ. ROOT)) .OR. &
                    ((INTER_FLAG .EQV. .TRUE.) .AND. &
                     (MPITEST_INTER .EQ. SEND_GROUP) .AND. &
                     (MPITEST_CURRENT_RANK .EQ. ROOT))) THEN


                  SEND_TO = COMM_SIZE
                  IF (INTER_FLAG) SEND_TO = RECEIVERS

                  DO 500 M=0, SEND_TO-1
                    IF ((INTER_FLAG .EQV. .FALSE.) .AND. &
                        (M .EQ. ROOT)) GO TO 500
!                       Don't send to self
                    LOOP_CNT = LOOP_CNT + 1
                    CALL MPITEST_DATATEMPLATE_INIT(ROOT_VALUE,  &
                      MPITEST_CURRENT_RANK, TEST_TYPE)
                    CALL MPITEST_INIT_BUFFER(TEST_TYPE,  &
                      ROOT_ALL_LENGTH, &
                      ROOT_VALUE, ROOT_ALL_BUFFER)


!=============================================================================
!                    ROOT SENDRECV_REPLACE
!-----------------------------------------------------------------------------
                    CALL MPI_SENDRECV_REPLACE( &
                             ROOT_ALL_BUFFER, &
                             ROOT_ALL_LENGTH, &
                             MPITEST_MPI_DATATYPES(TEST_TYPE), &
                             M,  &
                             ROOT, &
                             M, &
                             M,      &
                             COMM, &
                             ROOT_STAT,  &
                             IERR)

                    IF (IERR .NE. MPI_SUCCESS) THEN
                      INFOBUF=' '
                      WRITE(INFOBUF,99) &
                        'ROOT MPI_SENDRECV_REPLACE() returned', IERR
                      CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                      INFOBUF=' '
                      CALL MPI_ERROR_STRING(IERR, ERRORSTRING, ERRSIZE, &
                           ERR)
                      CALL MPITEST_MESSAGE(MPITEST_NONFATAL,ERRORSTRING)
                      FAIL = FAIL + 1
                    END IF
!
!                 Check data for errors
!
                  CALL MPITEST_DATATEMPLATE_INIT(ROOT_VALUE, M, &
                    TEST_TYPE)
                  CALL MPITEST_BUFFER_ERRORS(TEST_TYPE,  &
                    ROOT_ALL_LENGTH,  &
                    ROOT_VALUE, ROOT_ALL_BUFFER, ERROR)
                  IF ((ERROR) .NE. 0) THEN
                    INFOBUF=' '
                    WRITE (INFOBUF,189) ERROR, &
                      ' errors in buffer, ', &
                      'commindex',COMM_INDEX,', commsize', COMM_SIZE, &
                      ', commtype', COMM_TYPE, ', root', ROOT, &
                      ', datatype', TEST_TYPE, ', length',   &
                      ROOT_ALL_LENGTH
                    CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                    LOOP_FAIL = LOOP_FAIL + 1
                  END IF
                
!
!                 Call MPI_GET_COUNT and compare value with length received
!
                  CALL MPI_GET_COUNT(ROOT_STAT,  &
                      MPITEST_MPI_DATATYPES(TEST_TYPE), ROOT_CNT, IERR)
                  IF (IERR .NE. MPI_SUCCESS) THEN
                    INFOBUF=' '
                    WRITE(INFOBUF,99)  &
                      'ROOT: MPI_GET_COUNT() returned', IERR
                    CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                    INFOBUF=' '
                    CALL MPI_ERROR_STRING(IERR, ERRORSTRING, ERRSIZE, &
                         ERR)
                    CALL MPITEST_MESSAGE(MPITEST_NONFATAL, ERRORSTRING)
                    LOOP_FAIL = LOOP_FAIL + 1
                  END IF

                  IF (ROOT_ALL_LENGTH - ROOT_CNT .NE. 0) THEN
                    INFOBUF=' '
                    WRITE(INFOBUF,98) 'Unexpected receive length', &
                        ROOT_CNT, ', expected', ROOT_ALL_LENGTH
                    CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                    INFOBUF=' '
                    LOOP_FAIL = LOOP_FAIL + 1
                  END IF

!
!                 Check source and tag values.
!
                  IF (ROOT_STAT(MPI_TAG) .NE.  &
                                      ROOT_STAT(MPI_SOURCE) ) THEN
                    INFOBUF=' '
                    WRITE(INFOBUF,98) 'Unexpected tag value', &
                      ROOT_STAT(MPI_TAG), ', expected',  &
                      ROOT_STAT(MPI_SOURCE)
                    CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                    INFOBUF=' '
                    LOOP_FAIL = LOOP_FAIL + 1
                  END IF


                  IF (LOOP_FAIL .NE. 0) FAIL = FAIL + 1


 500              CONTINUE
!                         Send to all receivers
                END IF
!                    Sender

              CALL MPI_BARRIER(BARRIER_COMM, IERR)
              IF (IERR .NE. MPI_SUCCESS) THEN
                INFOBUF=' '
                WRITE(INFOBUF,99) 'MPI_BARRIER() returned', IERR
                CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
                INFOBUF=' '
                CALL MPI_ERROR_STRING(IERR, ERRORSTRING, ERRSIZE, ERR)
                CALL MPITEST_MESSAGE(MPITEST_NONFATAL, ERRORSTRING)
                FAIL = FAIL + 1
              END IF

 400          CONTINUE
!                     End root loop


 350        CONTINUE
!                   End INTER send group loop
      
 300      CONTINUE
!                 End data length loop
      
 200    CONTINUE
!               End data type loop

        IF (INTER_FLAG) THEN
          CALL MPI_COMM_FREE(BARRIER_COMM, IERR)
          IF (IERR .NE. MPI_SUCCESS) THEN
            INFOBUF=' '
            WRITE(INFOBUF,99) 'MPI_COMM_FREE() returned', IERR
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
            INFOBUF=' '
            CALL MPI_ERROR_STRING(IERR, ERRORSTRING, ERRSIZE, ERR)
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, ERRORSTRING)
            FAIL = FAIL + 1
          END IF
        END IF

 201    CONTINUE

        CALL MPITEST_FREE_COMMUNICATOR(COMM_TYPE, COMM, IERR)
        CALL MPI_BARRIER(MPI_COMM_WORLD, IERR)

 100  CONTINUE
!             End communicator loop

!
!     Report overall results
!
      CALL MPITEST_REPORT(LOOP_CNT - FAIL, FAIL, 0, TESTNAME)
      
      CALL MPI_FINALIZE(IERR)

      END
