! ----------------------------------------------------------------------
! MESSAGE PASSING INTERFACE TEST CASE SUITE
! 
! Copyright - 1996 Intel Corporation
! 
! Intel Corporation hereby grants a non-exclusive license under Intel's
! copyright to copy, modify and distribute this software for any purpose 
! and without fee, provided that the above copyright notice and the
! following paragraphs appear on all copies.
! 
! Intel Corporation makes no representation that the test cases
! comprising this suite are correct or are an accurate representation
! of any standard.
! 
! IN NO EVENT SHALL INTEL HAVE ANY LIABILITY FOR ANY DIRECT, INDIRECT
! OR SPECULATIVE DAMAGES, (INCLUDING WITHOUT LIMITING THE FOREGOING,
! CONSEQUENTIAL, INCIDENTAL AND SPECIAL DAMAGES) INCLUDING, BUT NOT
! LIMITED TO INFRINGEMENT, LOSS OF USE, BUSINESS INTERRUPTIONS, AND
! LOSS OF PROFITS, IRRESPECTIVE OF WHETHER INTEL HAS ADVANCE NOTICE OF
! THE POSSIBILITY OF ANY SUCH DAMAGES.
! 
! INTEL CORPORATION SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING,
! BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
! FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  THE SOFTWARE PROVIDED
! HEREUNDER IS ON AN "AS IS" BASIS AND INTEL CORPORATION HAS NO
! OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS OR
! MODIFICATIONS.
! 
! ----------------------------------------------------------------------
! **********************************************************************
!
!                          MPI_CART_SUB
!
! Reference:  MPI Standard,  Chapter 6:  Process Topologies
!             Section 6.5.6:  Partitioning of Cartesian structures
!
! Uses only intra-communicators,  all inter-communicators are skipped
! Must have a minimum of MINNODE nodes in the communicator
! The test program is limited to testing not more than MAXNDIMS dimensions
!
!
! **********************************************************************

#include "f90options.h"

      INCLUDE 'mpitest_cfgf90.h'

      PROGRAM MAIN
        USE MPI

      INCLUDE 'mpitestf90.h'
      INCLUDE 'externalf90.h'

!
      INTEGER    MPITEST_MAXDIMS
      PARAMETER  ( MPITEST_MAXDIMS = 4)
!                                      Maximum dimensions to generate
      INTEGER    MPITEST_MINNODE
      PARAMETER  ( MPITEST_MINNODE = 6)
!                                      Mimum nodes required for test
      CHARACTER  TEST_DESC*(*)
      PARAMETER( TEST_DESC = 'MPI_CART_SUB')
!
      INTEGER CNUMBR
!                           index for communications loop
      INTEGER COMM_INDEX
!                           the array index of the current comm
      INTEGER COMM_TYPE
!                           the index of the current communicator type
      INTEGER COMM_COUNT
!                           loop counter for communicator loop
      INTEGER COMM_SIZE
!                           number of ranks in the current communicator
      INTEGER LOOP_CNT
!                           counts total number of failures, loops
      INTEGER COMM
!                           MPI communicator under test
      INTEGER ERRSIZE
!                           length of error message
      INTEGER PASS
      INTEGER FAIL              
!                           counts total number of PASSes/FAILures/VERIFY
      INTEGER IERR
      INTEGER ERR
!                           return value from MPI calls
      INTEGER ERRORCLASS
      INTEGER SIZE 
!                           error class of IERR
      INTEGER I, J, K, M
!                           general loop indices
      INTEGER NNODES
!                           number of nodes in Cartesian structure
      INTEGER NDIMS
!                           number of dimensions to generate coordinates for
      INTEGER CART_DIMS
!                           number of dimensions returned by CARTDIMS_GET
      INTEGER WORLD_NUMP
!                           the number of processors in MPI_COMM_WORLD  
      INTEGER CART_NUMP
!                           the number of processors in Cartesian communicator
      INTEGER RANK_CART
!                           rank in the Cartesian communicator
      INTEGER SUB_DIMS
!                           dimensions in sub-communicator
      INTEGER SUB_NUMP 
!                           ranks in sub-communicator
      INTEGER RANK_SUB
!                           rank in the sub-communicator
      INTEGER TEST_VAL
!                           communicaton test value
      INTEGER DIM_REM 
!                           remaining dimension count 
      LOGICAL INTER_FLAG
!                           is comm under test an INTER communicator
      LOGICAL REORDER
!                           flag telling if ranks can be reordered
      CHARACTER*(IOSIZE)  INFOBUF
!                           buffer for passing messages to MPITEST
      CHARACTER*(MPI_MAX_ERROR_STRING)  ERRORSTRING
!                           string describing error flag
      CHARACTER*32   TESTNAME
!                           name of this test
      INTEGER COMM_CART
!                           Cartesian communicator
      INTEGER COMM_SUB
!                           SUB communicator
      INTEGER DIM_SIZE(MPITEST_MAXDIMS)
!                           processes in each dimension
      INTEGER SUB_COORDS(MPITEST_MAXDIMS)
!                           periodic assignment flag vector
      LOGICAL SUB_PERIODS(MPITEST_MAXDIMS)
!                           periodic assignment flag vector
      INTEGER TEST_DIM_SIZE(MPITEST_MAXDIMS)
!                           sub comm computed dimension sizes
      INTEGER SUB_DIM_SIZE(MPITEST_MAXDIMS)
!                           sub comm actual dimension sizes
      LOGICAL REM_DIMS(MPITEST_MAXDIMS)
!                           remaining dimension flags for CART_SUB
      LOGICAL PERIODS(MPITEST_MAXDIMS)
!                           flag to allow periodic assignment
      INTEGER COORDS(MPITEST_MAXDIMS)	
!                           periodic assignment flag vector
      INTEGER ALL_RANKS(2,MAX_RANKS)
!                           output buffer for MPI_GATHER
      INTEGER RANKS(2)
!                           old and new rank for each node
!-----------------------------------------------------------------------------
 98   FORMAT(A,A)
 99   FORMAT(A,INT_FMT)
 100  FORMAT(A200)
 101  FORMAT(A, A, INT_FMT)
 102  FORMAT(A,INT_FMT, A,INT_FMT, A,INT_FMT, A,INT_FMT, A,INT_FMT)
 103  FORMAT(A, INT_FMT, A, A)
!
!     Initialize the MPI environment
!
      CALL MPI_INIT(IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF,99) 'MPI_INIT() returned', IERR
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
!
!     Initialize the MPITEST environment
!
      CALL MPITEST_INIT(IERR)
      TESTNAME = 'MPI_CART_SUB'
      IF (MPITEST_ME .EQ. 0) THEN
        INFOBUF = 'Starting test '//TESTNAME
        CALL MPITEST_MESSAGE(MPITEST_INFO0, INFOBUF)
      END IF
!
!     Get number of communicators
!
      CALL MPITEST_NUM_COMM_SIZES(COMM_COUNT)
!
!     Can't run if not at least  MPITEST_MINNODE  nodes  
!
      IF(MPITEST_NUMP .LT. MPITEST_MINNODE) THEN
        WRITE(INFOBUF,102)  &
          'Ranks =', MPITEST_NUMP, &
          ' MPI_CART_SUB needs at least ', MPITEST_MINNODE
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
      FAIL = 0
      LOOP_CNT = 0
      TEST_VAL = -1
!
!---------------------  Loop over Communicators  ------------------------
!
      DO 4000 CNUMBR = 1,COMM_COUNT
        CALL MPITEST_GET_COMM_INDEX(CNUMBR, COMM_INDEX)
        CALL MPITEST_GET_COMM_TYPE(CNUMBR, COMM_TYPE)
        CALL MPITEST_GET_COMMUNICATOR(COMM_TYPE, COMM_INDEX, COMM, &
              COMM_SIZE)
!
!       Skip to end of loop if this node is not part of current
!       communicator
        IF (COMM .EQ. MPI_COMM_NULL) GO TO 3400
!
!       Test for INTER-communicator
!
        CALL MPI_COMM_TEST_INTER(COMM, INTER_FLAG, IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          INFOBUF=' '
          WRITE(INFOBUF,99) 'MPI_COMM_TEST_INTER() returned', IERR
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          INFOBUF=' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, ERRSIZE, ERR)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, ERRORSTRING)
          FAIL = FAIL + 1
        END IF
        IF(INTER_FLAG  .OR.  COMM_SIZE .LT. MPITEST_MINNODE) GO TO 3000
        NDIMS = MPITEST_MAXDIMS
 1220   CONTINUE
        J = MOD(COMM_SIZE, NDIMS)
        IF( J .NE. 0) THEN
          NDIMS = NDIMS -1
          GO TO 1220
        END IF
        DO 1240 I =1,NDIMS
          DIM_SIZE(I) = 0
 1240   CONTINUE
!
!--------------------------  MPI_DIMS_CREATE  ------------------------
!
        NNODES = COMM_SIZE
        CALL MPI_DIMS_CREATE( NNODES, &
                              NDIMS, &
                              DIM_SIZE, &
                              IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          FAIL = FAIL + 1
          CALL MPI_ERROR_CLASS(IERR, ERRORCLASS)
          WRITE(INFOBUF,102) 'MPI_DIMS_CREATE ERROR (', IERR,  &
            ')    ERRORCLASS = ', ERRORCLASS
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
        END IF
        DO 280 I = 1, NDIMS
          PERIODS(I) = .TRUE.
 280    CONTINUE
!
!  Adjust dimensions, so last dimension is not 1
!
        I = NDIMS
 300    IF(DIM_SIZE(I) .LE. 1  .AND.  I .GE. 1) THEN
          DIM_SIZE(I) = 0
          I = I - 1
          NDIMS = NDIMS - 1
          GO TO 300
        END IF
        SUB_DIMS = NDIMS - 1      
!
!--------------------------  MPI_CART_CREATE  ------------------------
!
        REORDER = .TRUE.
        CALL   MPI_CART_CREATE(COMM, &
                               NDIMS, &
                               DIM_SIZE, &
                               PERIODS, &
                               REORDER, &
                               COMM_CART, &
                               IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          FAIL = FAIL + 1
          CALL MPI_ERROR_CLASS(IERR, ERRORCLASS, ERR)
          WRITE(INFOBUF,102) 'MPI_CART_CREATE ERROR (', IERR,  &
            ')    ERRORCLASS = ', ERRORCLASS
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
        END IF
!     First, the condition where no dimensions are delete
        LOOP_CNT = LOOP_CNT + 1
        DO 520 I = 1, NDIMS
          REM_DIMS(I) = .TRUE.
 520    CONTINUE
!
!    --------------------- MPI_CART_SUB   ----------------------
!
      CALL MPI_CART_SUB(COMM_CART, &
                        REM_DIMS, &
                        COMM_SUB, &
                        IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        CALL MPI_ERROR_CLASS(IERR, ERRORCLASS, ERR)
        WRITE(INFOBUF,102) 'MPI_CART_SUB  ERROR (', IERR,  &
             ')    ERRORCLASS = ', ERRORCLASS
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        ERRORSTRING = ' '
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
        WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
!
!    ------------------------- MPI_CART_GET   ------------------------
!
      LOOP_CNT = LOOP_CNT + 1
      CALL MPI_CART_GET(COMM_SUB, &
                        NDIMS, &
                        SUB_DIM_SIZE, &
                        SUB_PERIODS, &
                        SUB_COORDS, &
                        IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        CALL MPI_ERROR_CLASS(IERR, ERRORCLASS)
        WRITE(INFOBUF,102) 'MPI_CART_GET ERROR (', IERR,  &
             ')    ERRORCLASS = ', ERRORCLASS
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        ERRORSTRING = ' '
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
        WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
      CALL MPI_COMM_RANK(COMM_SUB,  RANK_SUB, IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        CALL MPI_ERROR_CLASS(IERR, ERRORCLASS)
        WRITE(INFOBUF,102) 'MPI_COMM_RANK Error (', IERR,  &
          ')    ERRORCLASS = ', ERRORCLASS
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        ERRORSTRING = ' '
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
        WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
!
      IF(RANK_SUB .EQ. 0) TEST_VAL = 1
!
!  Broadcast from node zero of sub-comminicator
!
      CALL MPI_BCAST(TEST_VAL, 1, MPI_INTEGER, 0, COMM_SUB, IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        CALL MPI_ERROR_CLASS(IERR, ERRORCLASS, ERR)
        WRITE(INFOBUF,102) 'MPI_BCAST ERROR (', IERR,  &
             ')    ERRORCLASS = ', ERRORCLASS
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        ERRORSTRING = ' '
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
        WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
        FAIL = FAIL + 1
      END IF
!
!  Nodes in communicator check for broadcast value
!
      IF(TEST_VAL .NE. 1) THEN
        WRITE(INFOBUF,102) &
          'ERROR in broadcast value:  Expected 1,  Actual', TEST_VAL
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
        FAIL = FAIL + 1
        END IF
        CALL MPI_COMM_FREE(COMM_SUB, IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          FAIL = FAIL + 1
          CALL MPI_ERROR_CLASS(IERR, ERRORCLASS, ERR)
          WRITE(INFOBUF,102) 'MPI_COMM_FREE ERROR (', IERR,  &
            ')    ERRORCLASS = ', ERRORCLASS
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
        END IF
        IF(NDIMS .LE. 1) GO TO 2000
        TEST_VAL = -1
        LOOP_CNT = LOOP_CNT + 1
        DO 800 J = 1, NDIMS
          DIM_REM = DIM_SIZE(J)
          DO 620 K = 1, NDIMS
            REM_DIMS(K) = .TRUE.
 620      CONTINUE
          REM_DIMS(J) = .FALSE.
          M = 0
          DO 660 K = 1, NDIMS
            IF(K .NE. J) THEN
              TEST_DIM_SIZE(M) = DIM_SIZE(K)
              M = M + 1
            END IF
 660      CONTINUE
!
!    --------------------- MPI_CART_SUB   ----------------------
!
          CALL MPI_CART_SUB(COMM_CART, &
                            REM_DIMS, &
                            COMM_SUB, &
                            IERR)
          IF (IERR .NE. MPI_SUCCESS) THEN
            FAIL = FAIL + 1
            CALL MPI_ERROR_CLASS(IERR, ERRORCLASS, ERR)
            WRITE(INFOBUF,102) 'MPI_CART_SUB  ERROR (', IERR,  &
              ')    ERRORCLASS = ', ERRORCLASS
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
            ERRORSTRING = ' '
            CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
            WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
            CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
          END IF
!
!    ------------------------- MPI_CART_GET   ------------------------
!
          LOOP_CNT = LOOP_CNT + 1
          CALL MPI_CART_GET(COMM_SUB, &
                             SUB_DIMS, &
                             SUB_DIM_SIZE, &
                             SUB_PERIODS, &
                             SUB_COORDS, &
                             IERR)
          IF (IERR .NE. MPI_SUCCESS) THEN
            FAIL = FAIL + 1
            CALL MPI_ERROR_CLASS(IERR, ERRORCLASS)
            WRITE(INFOBUF,102) 'MPI_CART_GET ERROR (', IERR,  &
              ')    ERRORCLASS = ', ERRORCLASS
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
            ERRORSTRING = ' '
            CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
            WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
            CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
          END IF
          CALL MPI_COMM_RANK(COMM_SUB,  RANK_SUB, IERR)
          IF (IERR .NE. MPI_SUCCESS) THEN
            FAIL = FAIL + 1
            CALL MPI_ERROR_CLASS(IERR, ERRORCLASS)
            WRITE(INFOBUF,102) 'MPI_COMM_RANK Error (', IERR,  &
              ')    ERRORCLASS = ', ERRORCLASS
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
            ERRORSTRING = ' '
            CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
            WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
            CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
          END IF
          IF(RANK_SUB .EQ. 0) TEST_VAL = 1
!
!  Broadcast from node zero of sub-comminicator
!
          CALL MPI_BCAST(TEST_VAL, 1, MPI_INTEGER, 0, COMM_SUB, IERR)
          IF (IERR .NE. MPI_SUCCESS) THEN
            FAIL = FAIL + 1
            CALL MPI_ERROR_CLASS(IERR, ERRORCLASS, ERR)
            WRITE(INFOBUF,102) 'MPI_BCAST ERROR (', IERR,  &
             ')    ERRORCLASS = ', ERRORCLASS
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
            ERRORSTRING = ' '
            CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
            WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
            CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
            FAIL = FAIL + 1
          END IF
!
!  Nodes in communicator check for broadcast value
!
          IF(TEST_VAL .NE. 1) THEN
            WRITE(INFOBUF,102) &
              'ERROR in broadcast value:  Expected 1,  Actual', TEST_VAL
            CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
            FAIL = FAIL + 1
          END IF
          CALL MPI_COMM_FREE(COMM_SUB, IERR)
          IF (IERR .NE. MPI_SUCCESS) THEN
            FAIL = FAIL + 1
            CALL MPI_ERROR_CLASS(IERR, ERRORCLASS, ERR)
            WRITE(INFOBUF,102) 'MPI_COMM_FREE ERROR (', IERR,  &
              ')    ERRORCLASS = ', ERRORCLASS
            CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
            ERRORSTRING = ' '
            CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
            WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
            CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
          END IF
 800    CONTINUE
 2000   CONTINUE
        CALL MPI_COMM_FREE(COMM_CART, IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          FAIL = FAIL + 1
          CALL MPI_ERROR_CLASS(IERR, ERRORCLASS, ERR)
          WRITE(INFOBUF,102) 'MPI_COMM_FREE ERROR (', IERR,  &
             ')    ERRORCLASS = ', ERRORCLASS
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
        END IF
 3000   CONTINUE
        IF(INTER_FLAG) THEN
          WRITE(INFOBUF,102) 'Skipped inter-communicator'
          CALL MPITEST_MESSAGE(MPITEST_INFO1, INFOBUF)
        END IF
        IF(COMM_SIZE .LT. MPITEST_MINNODE) THEN
          WRITE(INFOBUF,102) 'Skipping:  Communicator size = ', &
            COMM_SIZE, ' Minimum size = ', MPITEST_MINNODE
          CALL MPITEST_MESSAGE(MPITEST_INFO1, INFOBUF)
        END IF
 3400   CONTINUE
!                Node defined for this communicator
        CALL MPI_BARRIER(MPI_COMM_WORLD, IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          FAIL = FAIL + 1
          CALL MPI_ERROR_CLASS(IERR, ERRORCLASS)
          WRITE(INFOBUF,102) 'MPI_BARRIER Error (', IERR,  &
             ')    ERRORCLASS = ', ERRORCLASS
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
        END IF
        CALL MPITEST_FREE_COMMUNICATOR(COMM_TYPE, COMM, IERR)
 4000 CONTINUE
!             End communicator loop
!
!     Report overall results
!
      CALL MPITEST_REPORT(LOOP_CNT - FAIL, FAIL, 0, TESTNAME)
      CALL MPI_FINALIZE(IERR)
      END
