! ----------------------------------------------------------------------
! MESSAGE PASSING INTERFACE TEST CASE SUITE
! 
! Copyright - 1996 Intel Corporation
! 
! Intel Corporation hereby grants a non-exclusive license under Intel's
! copyright to copy, modify and distribute this software for any purpose 
! and without fee, provided that the above copyright notice and the
! following paragraphs appear on all copies.
! 
! Intel Corporation makes no representation that the test cases
! comprising this suite are correct or are an accurate representation
! of any standard.
! 
! IN NO EVENT SHALL INTEL HAVE ANY LIABILITY FOR ANY DIRECT, INDIRECT
! OR SPECULATIVE DAMAGES, (INCLUDING WITHOUT LIMITING THE FOREGOING,
! CONSEQUENTIAL, INCIDENTAL AND SPECIAL DAMAGES) INCLUDING, BUT NOT
! LIMITED TO INFRINGEMENT, LOSS OF USE, BUSINESS INTERRUPTIONS, AND
! LOSS OF PROFITS, IRRESPECTIVE OF WHETHER INTEL HAS ADVANCE NOTICE OF
! THE POSSIBILITY OF ANY SUCH DAMAGES.
! 
! INTEL CORPORATION SPECIFICALLY DISCLAIMS ANY WARRANTIES INCLUDING,
! BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
! FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  THE SOFTWARE PROVIDED
! HEREUNDER IS ON AN "AS IS" BASIS AND INTEL CORPORATION HAS NO
! OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS OR
! MODIFICATIONS.
! 
! ----------------------------------------------------------------------
! **********************************************************************
!
!                          MPI_GRAPH_CREATE_noreorder
!
! Reference:  MPI Standard,  Chapter 6:  Process Topologies
!                            Section 6.5.2
!
! Uses only intra-communicators,  all inter-communicators are skipped
! Must have a minimum of MINNODE nodes in the communicator
! The test program is limited to testing not more than MAXNDIMS dimensions
!
! Model  Test with  reorder = FALSE
!
! **********************************************************************

#include "f90options.h"

      INCLUDE 'mpitest_cfgf90.h'

      PROGRAM MAIN
        USE MPI

      INCLUDE 'mpitestf90.h'
      INCLUDE 'externalf90.h'

!
      INTEGER  MPITEST_MAXDIMS
      PARAMETER  ( MPITEST_MAXDIMS = 4)
!                                      Maximum dimensions to generate
      INTEGER  MPITEST_MINNODE
      PARAMETER  ( MPITEST_MINNODE = 6)
!                                      Minimum nodes required by test
      INTEGER  MPITEST_MAXN
      PARAMETER  ( MPITEST_MAXN = 8)
!                                      Max number of neighbors per node
      CHARACTER  TEST_DESC*(*)
      PARAMETER  ( TEST_DESC = 'MPI_GRAPH_CREATE_noreorder')
!-----------------------------------------------------------------------------

      INTEGER  CNUMBR
!                           index for communications loop
      INTEGER  COMM_INDEX
!                           the array index of the current comm
      INTEGER  COMM_TYPE
!                           the index of the current communicator type
      INTEGER  COMM_COUNT
!                           loop counter for communicator loop
      INTEGER  COMM_SIZE
!                           number of ranks in the current communicator
      INTEGER  LOOP_CNT
!                           counts total number of failures, loops
      INTEGER  COMM
!                           MPI communicator under test
      INTEGER  ERRSIZE
!                           length of error message
      INTEGER PASS
      INTEGER FAIL              
!                           counts total number of PASSes/FAILures/VERIFY
      INTEGER IERR
      INTEGER ERR
!                           return value from MPI calls
      INTEGER ERRORCLASS
      INTEGER SIZE 
!                           error class of IERR
      INTEGER TEST_VAL
!                           test value to broadcast
      INTEGER I, J, K
!                           general loop indices
      INTEGER NNODES
!                           number of nodes in graph structure
      INTEGER WORLD_NUMP
!                           the number of processors in MPI_COMM_WORLD  
      INTEGER TEST_NUMP
!                           the number of processors in current communicator  
      INTEGER RANK_GRAPH
!                           rank in the graph communicator  
      LOGICAL INTER_FLAG
!                           Is comm under test an INTER communicator
      LOGICAL REORDER
!                           flag telling if ranks can be reordered
      CHARACTER*(IOSIZE)  INFOBUF
!                           buffer for passing messages to MPITEST
      CHARACTER*(MPI_MAX_ERROR_STRING)  ERRORSTRING
!                           string describing error flag
      CHARACTER*32   TESTNAME
!                           name of this test
      INTEGER COMM_GRAPH
!                           graph communicator
      INTEGER INDEX(MAX_RANKS)
!                           cumulative neighbor count
      INTEGER EDGES(MAX_RANKS * MPITEST_MAXN)
!                           node connections list
!-----------------------------------------------------------------------------
 98   FORMAT(A,A)
 99   FORMAT(A,INT_FMT)
 100  FORMAT(A200)
 101  FORMAT(A, A, INT_FMT)
 102  FORMAT(A, INT_FMT, A, INT_FMT, A)
 103  FORMAT(A, INT_FMT, A, A)
!
!     Initialize the MPI environment
!
      CALL MPI_INIT(IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        WRITE(INFOBUF,99) 'MPI_INIT() returned', IERR
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
!
!     Initialize the MPITEST environment
!
      CALL MPITEST_INIT(IERR)
      TESTNAME = 'MPI_GRAPH_CREATE_noreorder'
      IF (MPITEST_ME .EQ. 0) THEN
        INFOBUF = 'Starting test '//TESTNAME
        CALL MPITEST_MESSAGE(MPITEST_INFO0, INFOBUF)
      END IF
!
!     Get number of communicators
!
      CALL MPITEST_NUM_COMM_SIZES(COMM_COUNT)
!
!     Can't run if not at least  MPITEST_MINNODE  nodes
!
      CALL MPI_COMM_SIZE(MPI_COMM_WORLD, WORLD_NUMP, IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        CALL MPI_ERROR_CLASS(IERR, ERRORCLASS)
        WRITE(INFOBUF,102) 'MPI_COMM_SIZE Error (', IERR,  &
             ')    ERRORCLASS = ', ERRORCLASS
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        ERRORSTRING = ' '
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
        WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
      IF(WORLD_NUMP .LT. MPITEST_MINNODE) THEN
        WRITE(INFOBUF,102)  &
          'Ranks =', WORLD_NUMP, &
          ' MPI_GRAPH_CREATE needs at least ', MPITEST_MINNODE
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
      FAIL = 0
      LOOP_CNT = 0
      TEST_VAL = -1
!
!---------------------  Loop over Communicators  ------------------------
!
      DO 4000 CNUMBR = 1,COMM_COUNT
        CALL MPITEST_GET_COMM_INDEX(CNUMBR, COMM_INDEX)
        CALL MPITEST_GET_COMM_TYPE(CNUMBR, COMM_TYPE)
        CALL MPITEST_GET_COMMUNICATOR(COMM_TYPE, COMM_INDEX, COMM, &
          COMM_SIZE)
!
!       Skip to end of loop if this node is not part of current
!       communicator
        IF (COMM .EQ. MPI_COMM_NULL) GO TO 3400
!
!       Test for INTER-communicator
!
        CALL MPI_COMM_TEST_INTER(COMM, INTER_FLAG, IERR)
        IF (IERR .NE. MPI_SUCCESS) THEN
          INFOBUF=' '
          WRITE(INFOBUF,99) 'MPI_COMM_TEST_INTER() returned', IERR
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          INFOBUF=' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, ERRSIZE, ERR)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, ERRORSTRING)
          FAIL = FAIL + 1
        END IF
        IF(INTER_FLAG  .OR.  COMM_SIZE .LT. MPITEST_MINNODE) GO TO 3000
!
!  Decide how many dimensions to generate
!
!  --------------  Set up index and edge arrays  ---------------------
!
! assume a single string of nodes with wraparound 
!
        J = 3
!
! set up edges for node 0
!
        EDGES(1) = COMM_SIZE - 1
        EDGES(2) = 1
        DO  320 I = 1, COMM_SIZE     
          EDGES(J) = I - 1
          EDGES(J + 1) = I + 1
          J = J + 2
 320    CONTINUE 
!
!  set up edges for last node
!
        EDGES(2 * COMM_SIZE - 1) = COMM_SIZE - 2
        EDGES(2 * COMM_SIZE) = 0
      DO 340 I = 1, COMM_SIZE
        INDEX(I) = I * 2
  340 CONTINUE
!-----------------------------  MPI_Graph_create  --------------------------*/
      LOOP_CNT = LOOP_CNT + 1
      NNODES   = COMM_SIZE
      REORDER  = .FALSE.      
      CALL  MPI_GRAPH_CREATE(COMM, &
                             NNODES, &
                             INDEX, &
                             EDGES, &
                             REORDER, &
                             COMM_GRAPH, &
                             IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        CALL MPI_ERROR_CLASS(IERR, ERRORCLASS)
        WRITE(INFOBUF,102) 'MPI_GRAPH_CREATE ERROR (', IERR,  &
             ')    ERRORCLASS = ', ERRORCLASS

        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        ERRORSTRING = ' '
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
        WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
      CALL MPI_COMM_RANK(COMM_GRAPH,  RANK_GRAPH, IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        CALL MPI_ERROR_CLASS(IERR, ERRORCLASS)
          WRITE(INFOBUF,102) 'MPI_COMM_RANK Error (', IERR,  &
          ')    ERRORCLASS = ', ERRORCLASS
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
        END IF
      IF(RANK_GRAPH .EQ. 0) TEST_VAL = 1
!
!  Broadcast from node zero of graph comminicator
!
      CALL MPI_BCAST(TEST_VAL, 1, MPI_INTEGER, 0, COMM_GRAPH, IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        CALL MPI_ERROR_CLASS(IERR, ERRORCLASS)
        WRITE(INFOBUF,102) 'MPI_BCAST ERROR (', IERR,  &
             ')    ERRORCLASS = ', ERRORCLASS

        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        ERRORSTRING = ' '
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
        WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
        FAIL = FAIL + 1
      END IF
!
!  Nodes in communicator check for broadcast value
!
      IF(TEST_VAL .NE. 1) THEN
        WRITE(INFOBUF,102) &
          'ERROR in broadcast value:  Expected 1,  Actual', TEST_VAL
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
        FAIL = FAIL + 1
      END IF
!
!  Reorder is FALSE, check that  graph rank equals original rank
!
        IF (MPITEST_CURRENT_RANK .NE. RANK_GRAPH) THEN
          FAIL = FAIL + 1
          WRITE(INFOBUF,102) "Error:  Ranks reordered  Old/New = ", &
            MPITEST_CURRENT_RANK, "/",  RANK_GRAPH
          CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
        END IF
      CALL MPI_COMM_FREE(COMM_GRAPH, IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
        FAIL = FAIL + 1
        CALL MPI_ERROR_CLASS(IERR, ERRORCLASS)
        WRITE(INFOBUF,102) 'MPI_COMM_FREE ERROR (', IERR,  &
             ')    ERRORCLASS = ', ERRORCLASS
        CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
        ERRORSTRING = ' '
        CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
        WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
        CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
      END IF
 3000 CONTINUE
      IF(INTER_FLAG) THEN
        WRITE(INFOBUF,102) 'Skipped inter-communicator'
        CALL MPITEST_MESSAGE(MPITEST_INFO1, INFOBUF)
      END IF
      IF(COMM_SIZE .LT. MPITEST_MINNODE) THEN
        WRITE(INFOBUF,102) 'Skipping:  Communicator size = ', &
          COMM_SIZE, ' Minimum size = ', MPITEST_MINNODE
        CALL MPITEST_MESSAGE(MPITEST_INFO1, INFOBUF)
      END IF
 3400 CONTINUE
!
!     Node defined for this communicator
!
      CALL MPI_BARRIER(MPI_COMM_WORLD, IERR)
      IF (IERR .NE. MPI_SUCCESS) THEN
          FAIL = FAIL + 1
          CALL MPI_ERROR_CLASS(IERR, ERRORCLASS)
          WRITE(INFOBUF,102) 'MPI_BARRIER Error (', IERR,  &
             ')    ERRORCLASS = ', ERRORCLASS
          CALL MPITEST_MESSAGE(MPITEST_NONFATAL, INFOBUF)
          ERRORSTRING = ' '
          CALL MPI_ERROR_STRING(IERR, ERRORSTRING, SIZE, ERR)
          WRITE(INFOBUF,100) ERRORSTRING(1:SIZE)
          CALL MPITEST_MESSAGE(MPITEST_FATAL, INFOBUF)
        END IF
        CALL MPITEST_FREE_COMMUNICATOR(COMM_TYPE, COMM, IERR)
 4000 CONTINUE
!
!             End communicator loop
!
!     Report overall results
!
      CALL MPITEST_REPORT(LOOP_CNT - FAIL, FAIL, 0, TESTNAME)
      CALL MPI_FINALIZE(IERR)
      END
